#!/usr/bin/env python
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unittests for GPM IMERG reader."""

import datetime as dt
import os
import unittest
from unittest import mock

import dask.array as da
import h5py
import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

DEFAULT_FILE_SHAPE = (3600, 1800)
DEFAULT_LAT_DATA = np.linspace(-89.95, 89.95,
                               DEFAULT_FILE_SHAPE[1]).astype(np.float32)
DEFAULT_LON_DATA = np.linspace(-179.95, 179.95,
                               DEFAULT_FILE_SHAPE[0]).astype(np.float32)


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def _get_geo_data(self, num_rows, num_cols):
        geo = {
            "Grid/lon":
                xr.DataArray(DEFAULT_LON_DATA,
                             attrs={"units": "degrees_east", },
                             dims=("lon")),
            "Grid/lat":
                xr.DataArray(DEFAULT_LAT_DATA,
                             attrs={"units": "degrees_north", },
                             dims=("lat")),
        }
        return geo

    def _get_precip_data(self, num_rows, num_cols):
        selection = {
            "Grid/IRprecipitation":
            xr.DataArray(
                da.ones((1, num_cols, num_rows), chunks=1024,
                        dtype=np.float32),
                attrs={
                    "_FillValue": -9999.9,
                    "units": "mm/hr",
                    "Units": "mm/hr",
                    "badval": h5py.h5r.Reference(),
                    "badvals": np.array([[h5py.h5r.Reference()]])
                },
                dims=("time", "lon", "lat")),
        }
        return selection

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        num_rows = 1800
        num_cols = 3600

        test_content = {}
        data = {}
        data = self._get_geo_data(num_rows, num_cols)
        test_content.update(data)
        data = self._get_precip_data(num_rows, num_cols)
        test_content.update(data)

        return test_content


class TestHdf5IMERG(unittest.TestCase):
    """Test the GPM IMERG reader."""

    yaml_file = "gpm_imerg.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.gpm_imerg import Hdf5IMERG
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(Hdf5IMERG, "__bases__", (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_load_data(self):
        """Test loading data."""
        from pyproj import CRS

        from satpy.readers.core.loading import load_reader

        # Filename to test, needed for start and end times
        filenames = [
            "3B-HHR.MS.MRG.3IMERG.20200131-S233000-E235959.1410.V06B.HDF5", ]

        # Expected projection in area def
        pdict = {"proj": "longlat",
                 "datum": "WGS84",
                 "no_defs": None,
                 "type": "crs"}

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        assert 1 == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        res = reader.load(["IRprecipitation"])
        assert 1 == len(res)
        assert res["IRprecipitation"].start_time == dt.datetime(2020, 1, 31, 23, 30, 0)
        assert res["IRprecipitation"].end_time == dt.datetime(2020, 1, 31, 23, 59, 59)
        assert res["IRprecipitation"].resolution == 0.1
        assert res["IRprecipitation"].area.width == 3600
        assert res["IRprecipitation"].area.height == 1800
        assert res["IRprecipitation"].area.crs == CRS(pdict)
        np.testing.assert_almost_equal(res["IRprecipitation"].area.area_extent,
                                       (-179.95, -89.95, 179.95, 89.95), 5)



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Shared objects for writing image-like output."""
from __future__ import annotations

import typing

from satpy.enhancements.enhancer import get_enhanced_image
from satpy.writers.core.base import Writer

if typing.TYPE_CHECKING:
    from os import PathLike
    from typing import Any

    import dask.array as da
    import xarray as xr
    from dask.delayed import Delayed
    from trollimage.xrimage import XRImage


class ImageWriter(Writer):
    """Base writer for image file formats."""

    def __init__(self, name=None, filename=None, base_dir=None, enhance=None, **kwargs):
        """Initialize image writer object.

        Args:
            name (str): A name for this writer for log and error messages.
                If this writer is configured in a YAML file its name should
                match the name of the YAML file. Writer names may also appear
                in output file attributes.
            filename (str): Filename to save data to. This filename can and
                should specify certain python string formatting fields to
                differentiate between data written to the files. Any
                attributes provided by the ``.attrs`` of a DataArray object
                may be included. Format and conversion specifiers provided by
                the :class:`trollsift <trollsift.parser.StringFormatter>`
                package may also be used. Any directories in the provided
                pattern will be created if they do not exist. Example::

                    {platform_name}_{sensor}_{name}_{start_time:%Y%m%d_%H%M%S}.tif

            base_dir (str | pathlib.Path):
                Base destination directories for all created files.
            enhance (bool or Enhancer): Whether to automatically enhance
                data to be more visually useful and to fit inside the file
                format being saved to. By default, this will default to using
                the enhancement configuration files found using the default
                :class:`~satpy.enhancements.enhancer.Enhancer` class. This can be set to
                `False` so that no enhancements are performed. This can also
                be an instance of the :class:`~satpy.enhancements.enhancer.Enhancer` class
                if further custom enhancement is needed.

            kwargs (dict): Additional keyword arguments to pass to the
                :class:`~satpy.writers.core.base.Writer` base class.

        .. versionchanged:: 0.10

            Deprecated `enhancement_config_file` and 'enhancer' in favor of
            `enhance`. Pass an instance of the `Enhancer` class to `enhance`
            instead.

        """
        super().__init__(name, filename, base_dir, **kwargs)
        if enhance is False:
            # No enhancement
            self.enhancer = False
        elif enhance is None or enhance is True:
            # default enhancement
            from satpy.enhancements.enhancer import Enhancer

            enhancement_config = self.info.get("enhancement_config", None)
            self.enhancer = Enhancer(enhancement_config_file=enhancement_config)
        else:
            # custom enhancer
            self.enhancer = enhance

    @classmethod
    def separate_init_kwargs(cls, kwargs):
        """Separate the init kwargs."""
        # FUTURE: Don't pass Scene.save_datasets kwargs to init and here
        init_kwargs, kwargs = super(ImageWriter, cls).separate_init_kwargs(kwargs)
        for kw in ["enhancement_config", "enhance"]:
            if kw in kwargs:
                init_kwargs[kw] = kwargs.pop(kw)
        return init_kwargs, kwargs

    def save_dataset(
        self,
        dataset: xr.DataArray,
        filename: str | None = None,
        fill_value: float | int | None = None,
        compute: bool = True,
        units: str | None = None,
        overlay: dict | None = None,
        decorate: dict | None = None,
        **kwargs,
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | PathLike | None]:
        """Save the ``dataset`` to a given ``filename``.

        This method creates an enhanced image using :func:`~satpy.enhancements.enhancer.get_enhanced_image`.
        The image is then passed to :meth:`save_image`. See both of these
        functions for more details on the arguments passed to this method.

        """
        if units is not None:
            import pint_xarray  # noqa
            dataset = dataset.pint.quantify().pint.to(units).pint.dequantify()
        img = get_enhanced_image(dataset.squeeze(), enhance=self.enhancer, overlay=overlay,
                                 decorate=decorate, fill_value=fill_value)
        return self.save_image(img, filename=filename, compute=compute, fill_value=fill_value, **kwargs)

    def save_image(
            self,
            img: XRImage,
            filename: str | None = None,
            compute: bool = True,
            **kwargs
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | PathLike | None]:
        """Save Image object to a given ``filename``.

        Args:
            img (trollimage.xrimage.XRImage): Image object to save to disk.
            filename (str): Optionally specify the filename to save this
                            dataset to. It may include string formatting
                            patterns that will be filled in by dataset
                            attributes.
            compute (bool): If `True` (default), compute and save the dataset.
                            If `False` return either a :doc:`dask:delayed`
                            object or tuple of (source, target). See the
                            return values below for more information.
            **kwargs: Other keyword arguments to pass to this writer.

        Returns:
            Value returned depends on `compute`. If `compute` is `True` then
            the return value is the result of computing a
            :doc:`dask:delayed` object or running :func:`dask.array.store`.
            If `compute` is `False` then the returned value is either a
            :doc:`dask:delayed` object that can be computed using
            `delayed.compute()` or a tuple of (source, target) that should be
            passed to :func:`dask.array.store`. If target is provided the the
            caller is responsible for calling `target.close()` if the target
            has this method.

        """
        raise NotImplementedError("Writer '%s' has not implemented image saving" % (self.name,))



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utility functions for area definitions."""
from ._config import config_search_paths, get_config_path


def get_area_file():
    """Find area file(s) to use.

    The files are to be named `areas.yaml` or `areas.def`.
    """
    paths = config_search_paths("areas.yaml")
    if paths:
        return paths
    else:
        return get_config_path("areas.def")


def get_area_def(area_name):
    """Get the definition of *area_name* from file.

    The file is defined to use is to be placed in the $SATPY_CONFIG_PATH
    directory, and its name is defined in satpy's configuration file.
    """
    try:
        from pyresample import parse_area_file
    except ImportError:
        from pyresample.utils import parse_area_file
    return parse_area_file(get_area_file(), area_name)[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the HDF-EOS base functionality."""

import unittest

nrt_mda = """GROUP                  = INVENTORYMETADATA
  GROUPTYPE            = MASTERGROUP

  GROUP                  = ECSDATAGRANULE

    OBJECT                 = REPROCESSINGPLANNED
      NUM_VAL              = 1
      VALUE                = "further update is anticipated"
    END_OBJECT             = REPROCESSINGPLANNED

    OBJECT                 = REPROCESSINGACTUAL
      NUM_VAL              = 1
      VALUE                = "Near Real Time"
    END_OBJECT             = REPROCESSINGACTUAL

    OBJECT                 = LOCALGRANULEID
      NUM_VAL              = 1
      VALUE                = "MYD03.A2019051.1225.061.2019051131153.NRT.hdf"
    END_OBJECT             = LOCALGRANULEID

    OBJECT                 = PRODUCTIONDATETIME
      NUM_VAL              = 1
      VALUE                = "2019-02-20T13:11:53.000Z"
    END_OBJECT             = PRODUCTIONDATETIME

    OBJECT                 = DAYNIGHTFLAG
      NUM_VAL              = 1
      VALUE                = "Day"
    END_OBJECT             = DAYNIGHTFLAG

    OBJECT                 = LOCALVERSIONID
      NUM_VAL              = 1
      VALUE                = "6.0.4"
    END_OBJECT             = LOCALVERSIONID

  END_GROUP              = ECSDATAGRANULE

  GROUP                  = MEASUREDPARAMETER

    OBJECT                 = MEASUREDPARAMETERCONTAINER
      CLASS                = "1"

      OBJECT                 = PARAMETERNAME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "Geolocation"
      END_OBJECT             = PARAMETERNAME

      GROUP                  = QAFLAGS
        CLASS                = "1"

        OBJECT                 = AUTOMATICQUALITYFLAG
          NUM_VAL              = 1
          CLASS                = "1"
          VALUE                = "Passed"
        END_OBJECT             = AUTOMATICQUALITYFLAG

        OBJECT                 = AUTOMATICQUALITYFLAGEXPLANATION
          NUM_VAL              = 1
          CLASS                = "1"
          VALUE                = "Set to 'Failed' if processing error occurred, set to 'Passed' otherwise"
        END_OBJECT             = AUTOMATICQUALITYFLAGEXPLANATION

        OBJECT                 = SCIENCEQUALITYFLAG
          NUM_VAL              = 1
          VALUE                = "Not Investigated"
          CLASS                = "1"
        END_OBJECT             = SCIENCEQUALITYFLAG

      END_GROUP              = QAFLAGS

      GROUP                  = QASTATS
        CLASS                = "1"

        OBJECT                 = QAPERCENTMISSINGDATA
          NUM_VAL              = 1
          CLASS                = "1"
          VALUE                = 0
        END_OBJECT             = QAPERCENTMISSINGDATA

        OBJECT                 = QAPERCENTOUTOFBOUNDSDATA
          NUM_VAL              = 1
          CLASS                = "1"
          VALUE                = 0
        END_OBJECT             = QAPERCENTOUTOFBOUNDSDATA

      END_GROUP              = QASTATS

    END_OBJECT             = MEASUREDPARAMETERCONTAINER

  END_GROUP              = MEASUREDPARAMETER

  GROUP                  = ORBITCALCULATEDSPATIALDOMAIN

    OBJECT                 = ORBITCALCULATEDSPATIALDOMAINCONTAINER
      CLASS                = "1"

      OBJECT                 = ORBITNUMBER
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = 89393
      END_OBJECT             = ORBITNUMBER

      OBJECT                 = EQUATORCROSSINGLONGITUDE
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = -151.260740805733
      END_OBJECT             = EQUATORCROSSINGLONGITUDE

      OBJECT                 = EQUATORCROSSINGTIME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "12:49:52.965727"
      END_OBJECT             = EQUATORCROSSINGTIME

      OBJECT                 = EQUATORCROSSINGDATE
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "2019-02-20"
      END_OBJECT             = EQUATORCROSSINGDATE

    END_OBJECT             = ORBITCALCULATEDSPATIALDOMAINCONTAINER

  END_GROUP              = ORBITCALCULATEDSPATIALDOMAIN

  GROUP                  = COLLECTIONDESCRIPTIONCLASS

    OBJECT                 = SHORTNAME
      NUM_VAL              = 1
      VALUE                = "MYD03"
    END_OBJECT             = SHORTNAME

    OBJECT                 = VERSIONID
      NUM_VAL              = 1
      VALUE                = 61
    END_OBJECT             = VERSIONID

  END_GROUP              = COLLECTIONDESCRIPTIONCLASS

  GROUP                  = INPUTGRANULE

    OBJECT                 = INPUTPOINTER
      NUM_VAL              = 8
      VALUE                = ("MYD01.61.2019-051T12:25:00.000000Z.NA.29878844.500100_1.hdf", "MYD03LUT.coeff_V6.1.4", "PM1EPHND_NRT.A2019051.1220.061.2019051125628", "PM1EPHND_NRT.A2019051.1225.061.2019051125628", "PM1EPHND_NRT.A2019051.1230.061.2019051125628", "
          PM1ATTNR_NRT.A2019051.1220.061.2019051125628", "PM1ATTNR_NRT.A2019051.1225.061.2019051125628", "PM1ATTNR_NRT.A2019051.1230.061.2019051125628")
    END_OBJECT             = INPUTPOINTER

  END_GROUP              = INPUTGRANULE

  GROUP                  = SPATIALDOMAINCONTAINER

    GROUP                  = HORIZONTALSPATIALDOMAINCONTAINER

      GROUP                  = GPOLYGON

        OBJECT                 = GPOLYGONCONTAINER
          CLASS                = "1"

          GROUP                  = GRING
            CLASS                = "1"

            OBJECT                 = EXCLUSIONGRINGFLAG
              NUM_VAL              = 1
              CLASS                = "1"
              VALUE                = "N"
            END_OBJECT             = EXCLUSIONGRINGFLAG

          END_GROUP              = GRING

          GROUP                  = GRINGPOINT
            CLASS                = "1"

            OBJECT                 = GRINGPOINTLONGITUDE
              NUM_VAL              = 4
              CLASS                = "1"
              VALUE                = (25.3839329817764, 1.80418778807854, -6.50842421663422, 23.0260060198343)
            END_OBJECT             = GRINGPOINTLONGITUDE

            OBJECT                 = GRINGPOINTLATITUDE
              NUM_VAL              = 4
              CLASS                = "1"
              VALUE                = (29.5170117594673, 26.1480434828114, 43.2445462598877, 47.7959787025408)
            END_OBJECT             = GRINGPOINTLATITUDE

            OBJECT                 = GRINGPOINTSEQUENCENO
              NUM_VAL              = 4
              CLASS                = "1"
              VALUE                = (1, 2, 3, 4)
            END_OBJECT             = GRINGPOINTSEQUENCENO

          END_GROUP              = GRINGPOINT

        END_OBJECT             = GPOLYGONCONTAINER

      END_GROUP              = GPOLYGON

    END_GROUP              = HORIZONTALSPATIALDOMAINCONTAINER

  END_GROUP              = SPATIALDOMAINCONTAINER

  GROUP                  = RANGEDATETIME

    OBJECT                 = RANGEBEGINNINGTIME
      NUM_VAL              = 1
      VALUE                = "12:25:00.000000"
    END_OBJECT             = RANGEBEGINNINGTIME

    OBJECT                 = RANGEENDINGTIME
      NUM_VAL              = 1
      VALUE                = "12:30:00.000000"
    END_OBJECT             = RANGEENDINGTIME

    OBJECT                 = RANGEBEGINNINGDATE
      NUM_VAL              = 1
      VALUE                = "2019-02-20"
    END_OBJECT             = RANGEBEGINNINGDATE

    OBJECT                 = RANGEENDINGDATE
      NUM_VAL              = 1
      VALUE                = "2019-02-20"
    END_OBJECT             = RANGEENDINGDATE

  END_GROUP              = RANGEDATETIME

  GROUP                  = ASSOCIATEDPLATFORMINSTRUMENTSENSOR

    OBJECT                 = ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER
      CLASS                = "1"

      OBJECT                 = ASSOCIATEDSENSORSHORTNAME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "MODIS"
      END_OBJECT             = ASSOCIATEDSENSORSHORTNAME

      OBJECT                 = ASSOCIATEDPLATFORMSHORTNAME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "Aqua"
      END_OBJECT             = ASSOCIATEDPLATFORMSHORTNAME

      OBJECT                 = ASSOCIATEDINSTRUMENTSHORTNAME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "MODIS"
      END_OBJECT             = ASSOCIATEDINSTRUMENTSHORTNAME

    END_OBJECT             = ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER

  END_GROUP              = ASSOCIATEDPLATFORMINSTRUMENTSENSOR

  GROUP                  = PGEVERSIONCLASS

    OBJECT                 = PGEVERSION
      NUM_VAL              = 1
      VALUE                = "6.1.4"
    END_OBJECT             = PGEVERSION

  END_GROUP              = PGEVERSIONCLASS

  GROUP                  = ADDITIONALATTRIBUTES

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "1"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "1"
        NUM_VAL              = 1
        VALUE                = "GRANULENUMBER"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "1"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "1"
          VALUE                = "151"
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "2"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "2"
        NUM_VAL              = 1
        VALUE                = "SCI_STATE"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "2"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "2"
          VALUE                = "1"
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "3"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "3"
        NUM_VAL              = 1
        VALUE                = "SCI_ABNORM"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "3"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "3"
          VALUE                = "1"
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "5"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "5"
        NUM_VAL              = 1
        VALUE                = "PROCESSVERSION"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "5"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "5"
          VALUE                = "6.1.0"
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "4"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "4"
        NUM_VAL              = 1
        VALUE                = "GEO_EST_RMS_ERROR"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "4"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "4"
          VALUE                = "75      "
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "6"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "6"
        NUM_VAL              = 1
        VALUE                = "identifier_product_doi"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "6"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "6"
          VALUE                = "10.5067/MODIS/MYD03.NRT.061"
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

    OBJECT                 = ADDITIONALATTRIBUTESCONTAINER
      CLASS                = "7"

      OBJECT                 = ADDITIONALATTRIBUTENAME
        CLASS                = "7"
        NUM_VAL              = 1
        VALUE                = "identifier_product_doi_authority"
      END_OBJECT             = ADDITIONALATTRIBUTENAME

      GROUP                  = INFORMATIONCONTENT
        CLASS                = "7"

        OBJECT                 = PARAMETERVALUE
          NUM_VAL              = 1
          CLASS                = "7"
          VALUE                = "http://dx.doi.org"
        END_OBJECT             = PARAMETERVALUE

      END_GROUP              = INFORMATIONCONTENT

    END_OBJECT             = ADDITIONALATTRIBUTESCONTAINER

  END_GROUP              = ADDITIONALATTRIBUTES

END_GROUP              = INVENTORYMETADATA

END"""  # noqa: E501

nrt_mda_dict = {
    "INVENTORYMETADATA": {
        "ADDITIONALATTRIBUTES": {
            "ADDITIONALATTRIBUTESCONTAINER": {
                "ADDITIONALATTRIBUTENAME": {
                    "VALUE": "identifier_product_doi_authority"
                },
                "INFORMATIONCONTENT": {
                    "PARAMETERVALUE": {
                        "VALUE": "http://dx.doi.org"
                    }
                }
            }
        },
        "ASSOCIATEDPLATFORMINSTRUMENTSENSOR": {
            "ASSOCIATEDPLATFORMINSTRUMENTSENSORCONTAINER": {
                "ASSOCIATEDINSTRUMENTSHORTNAME": {
                    "VALUE": "MODIS"
                },
                "ASSOCIATEDPLATFORMSHORTNAME": {
                    "VALUE": "Aqua"
                },
                "ASSOCIATEDSENSORSHORTNAME": {
                    "VALUE": "MODIS"
                }
            }
        },
        "COLLECTIONDESCRIPTIONCLASS": {
            "SHORTNAME": {
                "VALUE": "MYD03"
            },
            "VERSIONID": {
                "VALUE": 61
            }
        },
        "ECSDATAGRANULE": {
            "DAYNIGHTFLAG": {
                "VALUE": "Day"
            },
            "LOCALGRANULEID": {
                "VALUE": "MYD03.A2019051.1225.061.2019051131153.NRT.hdf"
            },
            "LOCALVERSIONID": {
                "VALUE": "6.0.4"
            },
            "PRODUCTIONDATETIME": {
                "VALUE": "2019-02-20T13:11:53.000Z"
            },
            "REPROCESSINGACTUAL": {
                "VALUE": "Near "
                "Real "
                "Time"
            },
            "REPROCESSINGPLANNED": {
                "VALUE": "further "
                "update "
                "is "
                "anticipated"
            }
        },
        "GROUPTYPE": "MASTERGROUP",
        "INPUTGRANULE": {
            "INPUTPOINTER": {
                "VALUE":
                ("MYD01.61.2019-051T12:25:00.000000Z.NA.29878844.500100_1.hdf",
                 "MYD03LUT.coeff_V6.1.4",
                 "PM1EPHND_NRT.A2019051.1220.061.2019051125628",
                 "PM1EPHND_NRT.A2019051.1225.061.2019051125628",
                 "PM1EPHND_NRT.A2019051.1230.061.2019051125628", "          "
                 "PM1ATTNR_NRT.A2019051.1220.061.2019051125628",
                 "PM1ATTNR_NRT.A2019051.1225.061.2019051125628",
                 "PM1ATTNR_NRT.A2019051.1230.061.2019051125628")
            }
        },
        "MEASUREDPARAMETER": {
            "MEASUREDPARAMETERCONTAINER": {
                "PARAMETERNAME": {
                    "VALUE": "Geolocation"
                },
                "QAFLAGS": {
                    "AUTOMATICQUALITYFLAG": {
                        "VALUE": "Passed"
                    },
                    "AUTOMATICQUALITYFLAGEXPLANATION": {
                        "VALUE":
                        "Set "
                        "to "
                        "'Failed' "
                        "if "
                        "processing "
                        "error "
                        "occurred, "
                        "set "
                        "to "
                        "'Passed' "
                        "otherwise"
                    },
                    "SCIENCEQUALITYFLAG": {
                        "VALUE": "Not "
                        "Investigated"
                    }
                },
                "QASTATS": {
                    "QAPERCENTMISSINGDATA": {
                        "VALUE": 0
                    },
                    "QAPERCENTOUTOFBOUNDSDATA": {
                        "VALUE": 0
                    }
                }
            }
        },
        "ORBITCALCULATEDSPATIALDOMAIN": {
            "ORBITCALCULATEDSPATIALDOMAINCONTAINER": {
                "EQUATORCROSSINGDATE": {
                    "VALUE": "2019-02-20"
                },
                "EQUATORCROSSINGLONGITUDE": {
                    "VALUE": -151.260740805733
                },
                "EQUATORCROSSINGTIME": {
                    "VALUE": "12:49:52.965727"
                },
                "ORBITNUMBER": {
                    "VALUE": 89393
                }
            }
        },
        "PGEVERSIONCLASS": {
            "PGEVERSION": {
                "VALUE": "6.1.4"
            }
        },
        "RANGEDATETIME": {
            "RANGEBEGINNINGDATE": {
                "VALUE": "2019-02-20"
            },
            "RANGEBEGINNINGTIME": {
                "VALUE": "12:25:00.000000"
            },
            "RANGEENDINGDATE": {
                "VALUE": "2019-02-20"
            },
            "RANGEENDINGTIME": {
                "VALUE": "12:30:00.000000"
            }
        },
        "SPATIALDOMAINCONTAINER": {
            "HORIZONTALSPATIALDOMAINCONTAINER": {
                "GPOLYGON": {
                    "GPOLYGONCONTAINER": {
                        "GRING": {
                            "EXCLUSIONGRINGFLAG": {
                                "VALUE": "N"
                            }
                        },
                        "GRINGPOINT": {
                            "GRINGPOINTLATITUDE": {
                                "VALUE": (29.5170117594673, 26.1480434828114,
                                          43.2445462598877, 47.7959787025408)
                            },
                            "GRINGPOINTLONGITUDE": {
                                "VALUE": (25.3839329817764, 1.80418778807854,
                                          -6.50842421663422, 23.0260060198343)
                            },
                            "GRINGPOINTSEQUENCENO": {
                                "VALUE": (1, 2, 3, 4)
                            }
                        }
                    }
                }
            }
        }
    }
}

metadata_modisl1b = """
GROUP=SwathStructure
    GROUP=SWATH_1
        SwathName="MODIS_SWATH_Type_L1B"
            GROUP=DimensionMap
            OBJECT=DimensionMap_1
                GeoDimension="2*nscans"
                DataDimension="10*nscans"
                Offset=2
                Increment=5
            END_OBJECT=DimensionMap_1
            OBJECT=DimensionMap_2
                GeoDimension="1KM_geo_dim"
                DataDimension="Max_EV_frames"
                Offset=2
                Increment=5
            END_OBJECT=DimensionMap_2
        END_GROUP=DimensionMap
    END_GROUP=SWATH_1
END_GROUP=SwathStructure
END
"""  # noqa: E501

metadata_modisl2 = """
GROUP=SwathStructure
    GROUP=SWATH_1
        SwathName="mod35"
        GROUP=DimensionMap
            OBJECT=DimensionMap_1
                GeoDimension="Cell_Across_Swath_5km"
                DataDimension="Cell_Across_Swath_1km"
                Offset=2
                Increment=5
            END_OBJECT=DimensionMap_1
            OBJECT=DimensionMap_2
                GeoDimension="Cell_Along_Swath_5km"
                DataDimension="Cell_Along_Swath_1km"
                Offset=2
                Increment=5
            END_OBJECT=DimensionMap_2
        END_GROUP=DimensionMap
        GROUP=IndexDimensionMap
        END_GROUP=IndexDimensionMap
    END_GROUP=SWATH_1
END_GROUP=SwathStructure
END
"""  # noqa: E501


class TestReadMDA(unittest.TestCase):
    """Test reading metadata."""

    def test_read_mda(self):
        """Test reading basic metadata."""
        from satpy.readers.core.hdfeos import HDFEOSBaseFileReader
        res = HDFEOSBaseFileReader.read_mda(nrt_mda)
        assert res == nrt_mda_dict

    def test_read_mda_geo_resolution(self):
        """Test reading geo resolution."""
        from satpy.readers.core.hdfeos import HDFEOSGeoReader
        resolution_l1b = HDFEOSGeoReader.read_geo_resolution(
            HDFEOSGeoReader.read_mda(metadata_modisl1b)
            )
        assert resolution_l1b == 1000
        resolution_l2 = HDFEOSGeoReader.read_geo_resolution(
            HDFEOSGeoReader.read_mda(metadata_modisl2)
        )
        assert resolution_l2 == 5000



# Copyright (c) 2009-2024 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for ISCCP-NG L1G data (https://cimss.ssec.wisc.edu/isccp-ng/)."""

import logging

import numpy as np
import xarray as xr
from pyresample import geometry

from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)


class IsccpngL1gFileHandler(BaseFileHandler):
    """Reader L1G ISCCP-NG data."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(IsccpngL1gFileHandler, self).__init__(
            filename, filename_info, filetype_info)

        self._start_time = filename_info["start_time"]
        self._end_time = None
        self.sensor = "multiple_sensors"
        self.filename_info = filename_info

    def tile_geolocation(self, data, key):
        """Get geolocation on full swath."""
        if key in "latitude":
            return xr.DataArray(np.tile(data.values[:, np.newaxis], (1, 7200)), dims=["y", "x"], attrs=data.attrs)
        if key in "longitude":
            return xr.DataArray(np.tile(data.values, (3600, 1)), dims=["y", "x"], attrs=data.attrs)
        return data

    def get_best_layer_of_data(self, data):
        """Get the layer with best data (= layer 0). There are two more layers with additional data."""
        if len(data.dims) == 4:
            data = data[0, 0, :, :]
        return data.squeeze(drop=True)

    def get_area_def(self, dsid):
        """Get area definition."""
        proj_dict = {
            "proj": "latlong",
            "datum": "WGS84",
        }
        area = geometry.AreaDefinition(
            "lat lon grid",
            "name_of_proj",
            "id_of_proj",
            proj_dict,
            7200,
            3600,
            np.asarray([-180, -90, 180, 90])
        )
        return area

    def modify_dims_and_coords(self, data):
        """Remove coords and rename dims to x and y."""
        if len(data.dims) > 2:
            data = data.drop_vars("latitude")
            data = data.drop_vars("longitude")
            data = data.drop_vars("start_time")
            data = data.drop_vars("end_time")
            data = data.rename({"longitude": "x", "latitude": "y"})
        return data

    def set_time_attrs(self, data):
        """Set time from attributes."""
        if "start_time" in data.coords:
            data.attrs["start_time"] = data["start_time"].values[0]
            data.attrs["end_time"] = data["end_time"].values[0]
            self._end_time = data.attrs["end_time"]
            self._start_time = data.attrs["start_time"]

    def get_dataset(self, key, yaml_info):
        """Get dataset."""
        logger.debug("Getting data for: %s", yaml_info["name"])
        nc = xr.open_dataset(self.filename, chunks={"y": "auto", "x": 900})
        name = yaml_info.get("nc_store_name", yaml_info["name"])
        file_key = yaml_info.get("nc_key", name)
        data = nc[file_key]
        self.set_time_attrs(data)
        data = self.modify_dims_and_coords(data)
        data = self.get_best_layer_of_data(data)
        data = self.tile_geolocation(data, file_key)
        return data

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The HRIT msg reader tests package."""

import datetime as dt
import os
import unittest
import warnings
import zipfile
from unittest import mock

import fsspec
import numpy as np
import pytest
import xarray as xr
from numpy import testing as npt
from pyproj import CRS

import satpy.tests.reader_tests.seviri_l1b_hrit_setup as setup
from satpy.readers.seviri_l1b_hrit import HRITMSGEpilogueFileHandler, HRITMSGFileHandler, HRITMSGPrologueFileHandler
from satpy.tests.reader_tests.test_seviri_base import ORBIT_POLYNOMIALS_INVALID
from satpy.tests.reader_tests.test_seviri_l1b_calibration import TestFileHandlerCalibrationBase
from satpy.tests.utils import RANDOM_GEN, assert_attrs_equal, make_dataid


class TestHRITMSGBase(unittest.TestCase):
    """Baseclass for SEVIRI HRIT reader tests."""

    def assert_attrs_equal(self, attrs, attrs_exp):
        """Assert equality of dataset attributes."""
        assert_attrs_equal(attrs, attrs_exp, tolerance=1e-4)


class TestHRITMSGFileHandlerHRV(TestHRITMSGBase):
    """Test the HRITFileHandler."""

    def setUp(self):
        """Set up the hrit file handler for testing HRV."""
        self.observation_start_time = dt.datetime(2006, 1, 1, 12, 15, 9, 304888)
        self.nlines = 464
        self.reader = setup.get_fake_file_handler(
            observation_start_time=self.observation_start_time,
            nlines=self.nlines,
            ncols=5568,
        )
        self.reader.mda.update({
            "segment_sequence_number": 18,
            "planned_start_segment_number": 1
        })
        self.reader.fill_hrv = True

    @mock.patch("satpy.readers.core.hrit.np.memmap")
    def test_read_hrv_band(self, memmap):
        """Test reading the hrv band."""
        nbits = self.reader.mda["number_of_bits_per_pixel"]
        memmap.return_value = RANDOM_GEN.integers(0, 256,
                                                  size=int((464 * 5568 * nbits) / 8),
                                                  dtype=np.uint8)
        res = self.reader.read_band("HRV", None)
        assert res.shape == (464, 5568)

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITFileHandler.get_dataset")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.calibrate")
    def test_get_dataset(self, calibrate, parent_get_dataset):
        """Test getting the hrv dataset."""
        key = make_dataid(name="HRV", calibration="reflectance")
        info = setup.get_fake_dataset_info()

        parent_get_dataset.return_value = mock.MagicMock()
        calibrate.return_value = xr.DataArray(data=np.zeros((464, 5568)), dims=("y", "x"))
        res = self.reader.get_dataset(key, info)
        assert res.shape == (464, 11136)

        # Test method calls
        parent_get_dataset.assert_called_with(key, info)
        calibrate.assert_called_with(parent_get_dataset(), key["calibration"])

        self.assert_attrs_equal(res.attrs, setup.get_attrs_exp())
        np.testing.assert_equal(
            res["acq_time"],
            setup.get_acq_time_exp(self.observation_start_time, self.nlines)
        )

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITFileHandler.get_dataset")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.calibrate")
    def test_get_dataset_non_fill(self, calibrate, parent_get_dataset):
        """Test getting a non-filled hrv dataset."""
        key = make_dataid(name="HRV", calibration="reflectance")
        key.name = "HRV"
        info = setup.get_fake_dataset_info()
        self.reader.fill_hrv = False
        parent_get_dataset.return_value = mock.MagicMock()
        calibrate.return_value = xr.DataArray(data=np.zeros((464, 5568)), dims=("y", "x"))
        res = self.reader.get_dataset(key, info)
        assert res.shape == (464, 5568)

        # Test method calls
        parent_get_dataset.assert_called_with(key, info)
        calibrate.assert_called_with(parent_get_dataset(), key["calibration"])

        self.assert_attrs_equal(res.attrs, setup.get_attrs_exp())
        np.testing.assert_equal(
            res["acq_time"],
            setup.get_acq_time_exp(self.observation_start_time, self.nlines)
        )

    def test_get_area_def(self):
        """Test getting the area def."""
        area = self.reader.get_area_def(make_dataid(name="HRV", resolution=1000))
        assert area.area_extent == (-45561979844414.07, -3720765401003.719, 45602912357076.38, 77771774058.38356)

        expected_crs = CRS(dict(a=6378169.0, b=6356583.8, h=35785831.0, lon_0=0.0, proj="geos", units="m"))
        assert expected_crs == area.crs

        self.reader.fill_hrv = False
        area = self.reader.get_area_def(make_dataid(name="HRV", resolution=1000))
        npt.assert_allclose(area.defs[0].area_extent,
                            (-22017598561055.01, -2926674655354.9604, 23564847539690.22, 77771774058.38356))
        npt.assert_allclose(area.defs[1].area_extent,
                            (-30793529275853.656, -3720765401003.719, 14788916824891.568, -2926674655354.9604))

        assert area.defs[0].area_id == "msg_seviri_fes_1km"
        assert area.defs[1].area_id == "msg_seviri_fes_1km"


class TestHRITMSGFileHandler(TestHRITMSGBase):
    """Test the HRITFileHandler."""

    def setUp(self):
        """Set up the hrit file handler for testing."""
        self.observation_start_time = dt.datetime(2006, 1, 1, 12, 15, 9, 304888)
        self.nlines = 464
        self.ncols = 3712
        self.projection_longitude = 9.5
        self.reader = setup.get_fake_file_handler(
            observation_start_time=self.observation_start_time,
            nlines=self.nlines,
            ncols=self.ncols,
            projection_longitude=self.projection_longitude
        )
        self.reader.mda.update({
            "segment_sequence_number": 18,
            "planned_start_segment_number": 1
        })

    def _get_fake_data(self):
        return xr.DataArray(
            data=np.zeros((self.nlines, self.ncols)),
            dims=("y", "x")
        )

    def test_get_area_def(self):
        """Test getting the area def."""
        area = self.reader.get_area_def(make_dataid(name="VIS006", resolution=3000))

        expected_crs = CRS(dict(a=6378169.0, b=6356583.8, h=35785831.0, lon_0=self.projection_longitude,
                                proj="geos", units="m"))
        assert area.crs == expected_crs

        assert area.area_extent == (-77771774058.38356, -3720765401003.719, 30310525626438.438, 77771774058.38356)

        # Data shifted by 1.5km to N-W
        self.reader.mda["offset_corrected"] = False
        area = self.reader.get_area_def(make_dataid(name="VIS006", resolution=3000))
        assert area.area_extent == (-77771772558.38356, -3720765402503.719, 30310525627938.438, 77771772558.38356)

        assert area.area_id == "msg_seviri_rss_3km"

    @mock.patch("satpy.readers.core.hrit.np.memmap")
    def test_read_band(self, memmap):
        """Test reading a band."""
        nbits = self.reader.mda["number_of_bits_per_pixel"]
        memmap.return_value = RANDOM_GEN.integers(0, 256,
                                                  size=int((464 * 3712 * nbits) / 8),
                                                  dtype=np.uint8)
        res = self.reader.read_band("VIS006", None)
        assert res.shape == (464, 3712)

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITFileHandler.get_dataset")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.calibrate")
    def test_get_dataset(self, calibrate, parent_get_dataset):
        """Test getting the dataset."""
        data = self._get_fake_data()
        parent_get_dataset.return_value = mock.MagicMock()
        calibrate.return_value = data

        key = make_dataid(name="VIS006", calibration="reflectance")
        info = setup.get_fake_dataset_info()
        res = self.reader.get_dataset(key, info)

        # Test method calls
        new_data = np.zeros_like(data.data).astype("float32")
        new_data[:, :] = np.nan
        expected = data.copy(data=new_data)

        expected["acq_time"] = (
            "y",
            setup.get_acq_time_exp(self.observation_start_time, self.nlines)
        )
        xr.testing.assert_equal(res, expected)
        self.assert_attrs_equal(
            res.attrs,
            setup.get_attrs_exp(self.projection_longitude)
        )
        # testing start/end time
        assert dt.datetime(2006, 1, 1, 12, 15, 9, 304888) == self.reader.observation_start_time
        assert dt.datetime(2006, 1, 1, 12, 15) == self.reader.start_time
        assert self.reader.start_time == self.reader.nominal_start_time

        assert dt.datetime(2006, 1, 1, 12, 27, 39) == self.reader.observation_end_time
        assert self.reader.end_time == self.reader.nominal_end_time
        assert dt.datetime(2006, 1, 1, 12, 30) == self.reader.end_time
        # test repeat cycle duration
        assert 15 == self.reader._repeat_cycle_duration
        # Change the reducescan scenario to test the repeat cycle duration handling
        self.reader.epilogue["ImageProductionStats"]["ActualScanningSummary"]["ReducedScan"] = 1
        assert 5 == self.reader._repeat_cycle_duration

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITFileHandler.get_dataset")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.calibrate")
    def test_get_dataset_without_masking_bad_scan_lines(self, calibrate, parent_get_dataset):
        """Test getting the dataset."""
        data = self._get_fake_data()
        parent_get_dataset.return_value = mock.MagicMock()
        calibrate.return_value = data

        key = make_dataid(name="VIS006", calibration="reflectance")
        info = setup.get_fake_dataset_info()
        self.reader.mask_bad_quality_scan_lines = False
        res = self.reader.get_dataset(key, info)

        # Test method calls
        expected = data.copy()
        expected["acq_time"] = (
            "y",
            setup.get_acq_time_exp(self.observation_start_time, self.nlines)
        )
        xr.testing.assert_equal(res, expected)
        self.assert_attrs_equal(
            res.attrs,
            setup.get_attrs_exp(self.projection_longitude)
        )

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITFileHandler.get_dataset")
    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.calibrate")
    def test_get_dataset_with_raw_metadata(self, calibrate, parent_get_dataset):
        """Test getting the dataset."""
        calibrate.return_value = self._get_fake_data()
        key = make_dataid(name="VIS006", calibration="reflectance")
        info = setup.get_fake_dataset_info()
        self.reader.include_raw_metadata = True
        res = self.reader.get_dataset(key, info)
        assert "raw_metadata" in res.attrs

    def test_get_raw_mda(self):
        """Test provision of raw metadata."""
        self.reader.mda = {"segment": 1, "loff": 123}
        self.reader.prologue_.reduce = lambda max_size: {"prologue": 1}
        self.reader.epilogue_.reduce = lambda max_size: {"epilogue": 1}
        expected = {"prologue": 1, "epilogue": 1, "segment": 1}
        assert self.reader._get_raw_mda() == expected

        # Make sure _get_raw_mda() doesn't modify the original dictionary
        assert "loff" in self.reader.mda

    def test_satpos_no_valid_orbit_polynomial(self):
        """Test satellite position if there is no valid orbit polynomial."""
        reader = setup.get_fake_file_handler(
            observation_start_time=self.observation_start_time,
            nlines=self.nlines,
            ncols=self.ncols,
            projection_longitude=self.projection_longitude,
            orbit_polynomials=ORBIT_POLYNOMIALS_INVALID
        )
        assert "satellite_actual_longitude" not in reader.mda["orbital_parameters"]


class TestHRITMSGPrologueFileHandler(unittest.TestCase):
    """Test the HRIT prologue file handler."""

    def setUp(self, *mocks):
        """Set up the test case."""
        fh = setup.get_fake_file_handler(
            observation_start_time=dt.datetime(2016, 3, 3, 0, 0),
            nlines=464,
            ncols=3712,
        )
        self.reader = fh.prologue_

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGPrologueFileHandler.read_prologue")
    @mock.patch("satpy.readers.core.hrit.HRITFileHandler.__init__", autospec=True)
    def test_extra_kwargs(self, init, *mocks):
        """Test whether the prologue file handler accepts extra keyword arguments."""

        def init_patched(self, *args, **kwargs):
            self.mda = {}

        init.side_effect = init_patched

        HRITMSGPrologueFileHandler(filename="dummy_prologue_filename",
                                   filename_info={"service": ""},
                                   filetype_info=None,
                                   ext_calib_coefs={},
                                   mda_max_array_size=123,
                                   calib_mode="nominal")

    @mock.patch("satpy.readers.seviri_l1b_hrit.utils.reduce_mda")
    def test_reduce(self, reduce_mda):
        """Test metadata reduction."""
        reduce_mda.return_value = "reduced"

        # Set buffer
        assert self.reader.reduce(123) == "reduced"

        # Read buffer
        assert self.reader.reduce(123) == "reduced"
        reduce_mda.assert_called_once()


class TestHRITMSGEpilogueFileHandler(unittest.TestCase):
    """Test the HRIT epilogue file handler."""

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGEpilogueFileHandler.read_epilogue")
    @mock.patch("satpy.readers.core.hrit.HRITFileHandler.__init__", autospec=True)
    def setUp(self, init, *mocks):
        """Set up the test case."""

        def init_patched(self, *args, **kwargs):
            self.mda = {}

        init.side_effect = init_patched

        self.reader = HRITMSGEpilogueFileHandler(filename="dummy_epilogue_filename",
                                                 filename_info={"service": ""},
                                                 filetype_info=None,
                                                 calib_mode="nominal")

    @mock.patch("satpy.readers.seviri_l1b_hrit.HRITMSGEpilogueFileHandler.read_epilogue")
    @mock.patch("satpy.readers.core.hrit.HRITFileHandler.__init__", autospec=True)
    def test_extra_kwargs(self, init, *mocks):
        """Test whether the epilogue file handler accepts extra keyword arguments."""

        def init_patched(self, *args, **kwargs):
            self.mda = {}

        init.side_effect = init_patched

        HRITMSGEpilogueFileHandler(filename="dummy_epilogue_filename",
                                   filename_info={"service": ""},
                                   filetype_info=None,
                                   ext_calib_coefs={},
                                   mda_max_array_size=123,
                                   calib_mode="nominal")

    @mock.patch("satpy.readers.seviri_l1b_hrit.utils.reduce_mda")
    def test_reduce(self, reduce_mda):
        """Test metadata reduction."""
        reduce_mda.return_value = "reduced"

        # Set buffer
        assert self.reader.reduce(123) == "reduced"
        reduce_mda.assert_called()

        # Read buffer
        reduce_mda.reset_mock()
        self.reader._reduced = "red"
        assert self.reader.reduce(123) == "red"
        reduce_mda.assert_not_called()


class TestHRITMSGCalibration(TestFileHandlerCalibrationBase):
    """Unit tests for calibration."""

    @pytest.fixture(name="file_handler")
    def file_handler(self):
        """Create a mocked file handler."""
        prolog = {
            "RadiometricProcessing": {
                "Level15ImageCalibration": {
                    "CalSlope": self.gains_nominal,
                    "CalOffset": self.offsets_nominal,
                },
                "MPEFCalFeedback": {
                    "GSICSCalCoeff": self.gains_gsics,
                    "GSICSOffsetCount": self.offsets_gsics,
                }
            },
            "ImageDescription": {
                "Level15ImageProduction": {
                    "PlannedChanProcessing": self.radiance_types
                }
            },
            "ImageAcquisition": {
                "PlannedAcquisitionTime": {
                    "TrueRepeatCycleStart": self.scan_time,
                    }
                }
        }
        epilog = {
            "ImageProductionStats": {
                "ActualScanningSummary": {
                    "ForwardScanStart": self.scan_time
                }
            }
        }
        mda = {
            "image_segment_line_quality": {
                "line_validity": np.array([3, 3]),
                "line_radiometric_quality": np.array([4, 4]),
                "line_geometric_quality": np.array([4, 4])
            },
        }

        with mock.patch(
            "satpy.readers.seviri_l1b_hrit.HRITMSGFileHandler.__init__",
            return_value=None
        ):
            fh = HRITMSGFileHandler()
            fh.platform_id = self.platform_id
            fh.mda = mda
            fh.prologue = prolog
            fh.epilogue = epilog
            return fh

    @pytest.mark.parametrize(
        ("channel", "calibration", "calib_mode", "use_ext_coefs"),
        [
            # VIS channel, internal coefficients
            ("VIS006", "counts", "NOMINAL", False),
            ("VIS006", "radiance", "NOMINAL", False),
            ("VIS006", "radiance", "GSICS", False),
            ("VIS006", "reflectance", "NOMINAL", False),
            # VIS channel, external coefficients (mode should have no effect)
            ("VIS006", "radiance", "GSICS", True),
            ("VIS006", "reflectance", "NOMINAL", True),
            # IR channel, internal coefficients
            ("IR_108", "counts", "NOMINAL", False),
            ("IR_108", "radiance", "NOMINAL", False),
            ("IR_108", "radiance", "GSICS", False),
            ("IR_108", "brightness_temperature", "NOMINAL", False),
            ("IR_108", "brightness_temperature", "GSICS", False),
            # IR channel, external coefficients (mode should have no effect)
            ("IR_108", "radiance", "NOMINAL", True),
            ("IR_108", "brightness_temperature", "GSICS", True),
            # HRV channel, internal coefficiens
            ("HRV", "counts", "NOMINAL", False),
            ("HRV", "radiance", "NOMINAL", False),
            ("HRV", "radiance", "GSICS", False),
            ("HRV", "reflectance", "NOMINAL", False),
            # HRV channel, external coefficients (mode should have no effect)
            ("HRV", "radiance", "GSICS", True),
            ("HRV", "reflectance", "NOMINAL", True),
        ]
    )
    def test_calibrate(
            self, file_handler, counts, channel, calibration, calib_mode,
            use_ext_coefs
    ):
        """Test the calibration."""
        external_coefs = self.external_coefs if use_ext_coefs else {}
        expected = self._get_expected(
            channel=channel,
            calibration=calibration,
            calib_mode=calib_mode,
            use_ext_coefs=use_ext_coefs
        )

        fh = file_handler
        fh.mda["spectral_channel_id"] = self.spectral_channel_ids[channel]
        fh.channel_name = channel
        fh.calib_mode = calib_mode
        fh.ext_calib_coefs = external_coefs
        res = fh.calibrate(counts, calibration)
        xr.testing.assert_allclose(res, expected)

    def test_mask_bad_quality(self, file_handler):
        """Test the masking of bad quality scan lines."""
        channel = "VIS006"
        expected = self._get_expected(
            channel=channel,
            calibration="radiance",
            calib_mode="NOMINAL",
            use_ext_coefs=False
        )

        fh = file_handler

        res = fh._mask_bad_quality(expected)
        new_data = np.zeros_like(expected.data).astype("float32")
        new_data[:, :] = np.nan
        expected = expected.copy(data=new_data)
        xr.testing.assert_equal(res, expected)


@pytest.fixture(scope="session")
def prologue_file(session_tmp_path, prologue_header_contents):
    """Create a dummy prologue file."""
    from satpy.readers.seviri_l1b_native_hdr import hrit_prologue
    header = prologue_header_contents
    contents = np.void(1, dtype=hrit_prologue)
    contents["SatelliteStatus"]["SatelliteDefinition"]["SatelliteId"] = 324
    return create_file(session_tmp_path / "prologue", header + [contents])


@pytest.fixture(scope="session")
def prologue_header_contents():
    """Get the contents of the header."""
    return [
        # prime header
        np.void((0, 16), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((128, 90, 3403688),
                dtype=[("file_type", "u1"), ("total_header_length", ">u4"), ("data_field_length", ">u8")]),
        # second header
        np.void((4, 64), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.array(b"H-000-MSG4__-MSG4________-_________-PRO______-201802281500-__", dtype="|S61"),
        # timestamp record
        np.void((5, 10), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((64, (21973, 54911033)),
                dtype=[("cds_p_field", "u1"), ("timestamp", [("Days", ">u2"), ("Milliseconds", ">u4")])])
    ]


@pytest.fixture(scope="session")
def epilogue_file(session_tmp_path, epilogue_header_contents):
    """Create a dummy epilogue file."""
    from satpy.readers.seviri_l1b_native_hdr import hrit_epilogue
    header = epilogue_header_contents
    contents = np.void(1, dtype=hrit_epilogue)
    return create_file(session_tmp_path / "epilogue", header + [contents])


@pytest.fixture(scope="session")
def epilogue_header_contents():
    """Get the contents of the header."""
    return [
        np.void((0, 16), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((129, 90, 3042600),
                dtype=[("file_type", "u1"), ("total_header_length", ">u4"), ("data_field_length", ">u8")]),
        np.void((4, 64), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.array(b"H-000-MSG4__-MSG4________-_________-EPI______-201802281500-__", dtype="|S61"),
        np.void((5, 10), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((64, (21973, 54911033)),
                dtype=[("cds_p_field", "u1"), ("timestamp", [("Days", ">u2"), ("Milliseconds", ">u4")])]),
    ]


def create_file(filename, file_contents):
    """Create an hrit file."""
    with open(filename, "wb") as fh:
        for array in file_contents:
            array.tofile(fh)
    return filename


@pytest.fixture(scope="session")
def segment_file(session_tmp_path):
    """Create a segment_file."""
    cols = 3712
    lines = 464
    bpp = 10
    header = [
        np.void((0, 16), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((0, 6198, 17223680), dtype=[("file_type", "u1"), ("total_header_length", ">u4"),
                                            ("data_field_length", ">u8")]),
        np.void((1, 9), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((bpp, cols, lines, 0), dtype=[("number_of_bits_per_pixel", "u1"), ("number_of_columns", ">u2"),
                                           ("number_of_lines", ">u2"), ("compression_flag_for_data", "u1")]),
        np.void((2, 51), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((b"GEOS(+000.0)                    ", -13642337, -13642337, 1856, 1856),
                dtype=[("projection_name", "S32"),
                       ("cfac", ">i4"), ("lfac", ">i4"),
                       ("coff", ">i4"), ("loff", ">i4")]),
        np.void((4, 64), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.array(b"H-000-MSG4__-MSG4________-VIS008___-000001___-201802281500-__", dtype="|S61"),
        np.void((5, 10), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((64, (21973, 54911033)), dtype=[("cds_p_field", "u1"), ("timestamp", [("Days", ">u2"),
                                                                                      ("Milliseconds", ">u4")])]),
        np.void((128, 13), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.void((324, 2, 1, 1, 8, 0), dtype=[("GP_SC_ID", ">i2"), ("spectral_channel_id", "i1"),
                                             ("segment_sequence_number", ">u2"),
                                             ("planned_start_segment_number", ">u2"),
                                             ("planned_end_segment_number", ">u2"),
                                             ("data_field_representation", "i1")]),
        np.void((129, 6035), dtype=[("hdr_id", "u1"), ("record_length", ">u2")]),
        np.zeros((464, ), dtype=[("line_number_in_grid", ">i4"),
                                 ("line_mean_acquisition", [("days", ">u2"), ("milliseconds", ">u4")]),
                                 ("line_validity", "u1"), ("line_radiometric_quality", "u1"),
                                 ("line_geometric_quality", "u1")]),
        ]
    contents = np.empty(cols * lines * bpp // 8, dtype="u1")

    return create_file(session_tmp_path / "segment", header + [contents])


def test_read_real_segment(prologue_file, epilogue_file, segment_file):
    """Test reading an hrit segment."""
    info = dict(start_time=dt.datetime(2018, 2, 28, 15, 0), service="")
    prologue_fh = HRITMSGPrologueFileHandler(prologue_file, info, dict())
    epilogue_fh = HRITMSGEpilogueFileHandler(epilogue_file, info, dict())
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="No orbit polynomial valid")
        filehandler = HRITMSGFileHandler(segment_file, info, dict(), prologue_fh, epilogue_fh)
    res = filehandler.get_dataset(dict(name="VIS008", calibration="counts"),
                                  dict(units="", wavelength=0.8, standard_name="counts"))
    res.compute()


@pytest.fixture(scope="session")
def compressed_seviri_hrit_files(session_tmp_path, prologue_file, epilogue_file, segment_file):
    """Return the fsspec paths to the given seviri hrit files inside a zip file."""
    zip_full_path = session_tmp_path / "test_seviri_hrit.zip"
    with zipfile.ZipFile(zip_full_path, mode="w") as archive:
        for filename in (prologue_file, epilogue_file, segment_file):
            archive.write(filename, os.path.basename(filename))
    return {hrit_file: f"zip://{hrit_file}::file://{zip_full_path.as_posix()}"
            for hrit_file in ("prologue", "epilogue", "segment")}

def test_read_real_segment_zipped(compressed_seviri_hrit_files):
    """Test reading a remote hrit segment passed as FSFile."""
    from satpy.readers.core.remote import FSFile

    info = dict(start_time=dt.datetime(2018, 2, 28, 15, 0), service="")
    prologue = FSFile(fsspec.open(compressed_seviri_hrit_files["prologue"]))
    prologue_fh = HRITMSGPrologueFileHandler(prologue, info, dict())
    epilogue = FSFile(fsspec.open(compressed_seviri_hrit_files["epilogue"]))
    epilogue_fh = HRITMSGEpilogueFileHandler(epilogue, info, dict())
    segment = FSFile(fsspec.open(compressed_seviri_hrit_files["segment"]))
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="No orbit polynomial valid")
        filehandler = HRITMSGFileHandler(segment, info, dict(), prologue_fh, epilogue_fh)
    res = filehandler.get_dataset(dict(name="VIS008", calibration="counts"),
                                  dict(units="", wavelength=0.8, standard_name="counts"))
    res.compute()


def to_upath(fsfile):
    """Convert FSFile instance to UPath."""
    from upath import UPath
    fsfile_fs = fsfile.fs.to_dict()
    fsfile_fs.pop("cls")
    path = UPath(os.fspath(fsfile), **fsfile_fs)
    return path


@pytest.mark.filterwarnings("ignore:UPath 'zip' filesystem not explicitly implemented:UserWarning")
def test_read_real_segment_zipped_with_upath(compressed_seviri_hrit_files):
    """Test reading a remote hrit segment passed as UPath."""
    from satpy.readers.core.remote import FSFile

    info = dict(start_time=dt.datetime(2018, 2, 28, 15, 0), service="")

    prologue = FSFile(fsspec.open(compressed_seviri_hrit_files["prologue"]))
    prologue = to_upath(prologue)
    prologue_fh = HRITMSGPrologueFileHandler(prologue, info, dict())

    epilogue = FSFile(fsspec.open(compressed_seviri_hrit_files["epilogue"]))
    epilogue = to_upath(epilogue)
    epilogue_fh = HRITMSGEpilogueFileHandler(epilogue, info, dict())

    segment = FSFile(fsspec.open(compressed_seviri_hrit_files["segment"]))
    segment = to_upath(segment)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="No orbit polynomial valid")
        filehandler = HRITMSGFileHandler(segment, info, dict(), prologue_fh, epilogue_fh)
    res = filehandler.get_dataset(dict(name="VIS008", calibration="counts"),
                                  dict(units="", wavelength=0.8, standard_name="counts"))
    res.compute()



#!/usr/bin/env python
# Copyright (c) 2018-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Enhancements specific to the VIIRS instrument."""
import numpy as np
from trollimage.colormap import Colormap

from satpy.enhancements.wrappers import exclude_alpha, using_map_blocks


def water_detection(img, **kwargs):
    """Palettizes images from VIIRS flood data.

    This modifies the image's data so the correct colors
    can be applied to it, and then palettizes the image.
    """
    palette = kwargs["palettes"]
    palette["colors"] = tuple(map(tuple, palette["colors"]))

    _water_detection(img.data)
    cm = Colormap(*palette["colors"])
    img.palettize(cm)


@exclude_alpha
@using_map_blocks
def _water_detection(img_data):
    data = np.asarray(img_data).copy()
    data[data == 150] = 31
    data[data == 199] = 18
    data[data >= 200] = data[data >= 200] - 100

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""SEVIRI L2 BUFR format reader.

References:
    EUMETSAT Product Navigator
    https://navigator.eumetsat.int/

"""

import datetime as dt
import logging
import os

import dask.array as da
import numpy as np
import xarray as xr

from satpy.area import get_area_def
from satpy.readers.core._geos_area import get_geos_area_naming
from satpy.readers.core.eum import get_service_mode, recarray2dict
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.seviri import mpef_product_header
from satpy.utils import get_legacy_chunk_size

try:
    import eccodes as ec
except ImportError:
    raise ImportError(
        "Missing eccodes-python and/or eccodes C-library installation. Use conda to install eccodes")

logger = logging.getLogger("EumetsatL2Bufr")

CHUNK_SIZE = get_legacy_chunk_size()

SSP_DEFAULT = 0.0
BUFR_FILL_VALUE = -1.e+100

data_center_dict = {55: {"ssp": "E0415", "name": "MSG1"}, 56: {"ssp": "E0455", "name": "MSG2"},
                    57: {"ssp": "E0095", "name": "MSG3"}, 70: {"ssp": "E0000", "name": "MSG4"},
                    71: {"ssp": "E0000", "name": "MTGi1"}}

# Sensor resolution (pixel size in m) used to deermine product segment sizes
resolution_dict = {"fci": 2000, "seviri": 3000}

# List of variables that are now returned by eccodes as array, but that we want as single value
deprecate_to_single_value = ["satelliteIdentifier"]


class EumetsatL2BufrFileHandler(BaseFileHandler):
    """File handler for EUMETSAT Central Facility SEVIRI and FCI L2 BUFR products.

    **Loading data with AreaDefinition**

    By providing the `with_area_definition` as True in the `reader_kwargs`, the dataset is loaded with
    an AreaDefinition using a standardized AreaDefinition in areas.yaml. By default, the dataset will
    be loaded with a SwathDefinition, i.e. similar to how the data are stored in the BUFR file:

        scene = satpy.Scene(filenames,
                            reader="seviri_l2_bufr",
                            reader_kwargs={"with_area_definition": False})

    **Defining dataset recticifation longitude**

    The BUFR data were originally extracted from a rectified two-dimensional grid with a given central longitude
    (typically the sub-satellite point). This information is not available in the file itself nor the filename (for
    files from the EUMETSAT archive). Also, it cannot be reliably derived from all datasets themselves. Hence, the
    rectification longitude can be defined by the user by providing `rectification_longitude` in the `reader_kwargs`:

        scene = satpy.Scene(filenames,
                            reader="seviri_l2_bufr",
                            reader_kwargs={"rectification_longitude": 0.0})

    If not done, default values applicable to the operational grids of the respective SEVIRI instruments will be used.
    """

    def __init__(self, filename, filename_info, filetype_info, with_area_definition=False,
                 rectification_longitude="default", **kwargs):
        """Initialise the file handler for EUMETSAT SEVIRI and FCI L2 BUFR data."""
        super(EumetsatL2BufrFileHandler, self).__init__(filename,
                                                        filename_info,
                                                        filetype_info)

        # Set this environment variable to get consistent array sizes from eccodes. This fixes the cases
        # where all values in the expected array are the same (in particular fill values) which causes
        # eccodes to encode them and return them as a single value
        os.environ["ECCODES_BUFR_MULTI_ELEMENT_CONSTANT_ARRAYS"] = "1"

        if ("server" in filename_info):
            # EUMETSAT Offline Bufr product
            self.bufr_header = self._read_mpef_header()
        else:
            # Product was retrieved from the EUMETSAT Data Center
            attr = self.get_attributes(["typicalDate", "typicalTime", "satelliteIdentifier"])
            timeStr = attr["typicalDate"]+attr["typicalTime"]
            sc_id = int(attr["satelliteIdentifier"])

            self.bufr_header = {}
            self.bufr_header["NominalTime"] = dt.datetime.strptime(timeStr, "%Y%m%d%H%M%S")
            self.bufr_header["SpacecraftName"] = data_center_dict[sc_id]["name"]
            self.bufr_header["RectificationLongitude"] = data_center_dict[sc_id]["ssp"]

        if rectification_longitude != "default":
            self.bufr_header["RectificationLongitude"] = f"E{int(rectification_longitude * 10):04d}"

        self.filetype = filetype_info["file_type"]
        self.with_adef = with_area_definition

    def __del__(self):
        """Delete the instance and environment variable."""
        try:
            del os.environ["ECCODES_BUFR_MULTI_ELEMENT_CONSTANT_ARRAYS"]
        except KeyError:
            pass

    @property
    def start_time(self):
        """Return the repeat cycle start time."""
        return self.bufr_header["NominalTime"]

    @property
    def platform_name(self):
        """Return spacecraft name."""
        if self.bufr_header["SpacecraftName"].isdigit():
            # self._read_mpef_header will return the SpacecraftName as a string with an integer, representing the
            # Meteosat series number (1-11). This is cnverted to the corresponding MSG ID.
            return f'MSG{int(self.bufr_header["SpacecraftName"])-7}'
        else:
            return self.bufr_header["SpacecraftName"]

    @property
    def sensor_name(self):
        """Return instrument name."""
        if self.platform_name.startswith("MSG"):
            return "seviri"
        elif self.platform_name.startswith("MTG"):
            return "fci"

    @property
    def ssp_lon(self):
        """Return subsatellite point longitude."""
        ssp_lon = self.bufr_header["RectificationLongitude"]
        return float(ssp_lon[1:])/10.

    def get_area_def(self, key):
        """Return the area definition."""
        try:
            return self._area_def
        except AttributeError:
            raise NotImplementedError

    def _read_mpef_header(self):
        """Read MPEF header."""
        hdr = np.fromfile(self.filename, mpef_product_header, 1)
        return recarray2dict(hdr)

    def get_attributes(self, keys):
        """Get BUFR attributes."""
        # This function is inefficient as it is looping through the entire
        # file to get a list of attributes. It causes a problem though if you break
        # from the file early - dont know why but investigating - fix later.
        fh = open(self.filename, "rb")

        # Initialize output
        attrs = dict()

        while True:
            # Get handle for message
            bufr = ec.codes_bufr_new_from_file(fh)
            if bufr is None:
                break
            ec.codes_set(bufr, "unpack", 1)
            for k in keys:
                try:
                    if k in deprecate_to_single_value:
                        # With ECCODES_BUFR_MULTI_ELEMENT_CONSTANT_ARRAYS set to 1 all values, including scalars, are
                        # returned as arrays. Hence, we extract the single value here.
                        value = ec.codes_get_array(bufr, k)[0]
                    else:
                        value = ec.codes_get(bufr, k)
                    attrs[k] = value

                except BaseException:
                    attrs[k] = None
                    logging.warning(f"Failed to read key {k} from message")

            ec.codes_release(bufr)

        fh.close()
        return attrs

    def get_array(self, key):
        """Get all data from file for the given BUFR key."""
        with open(self.filename, "rb") as fh:
            msgCount = 0
            while True:
                bufr = ec.codes_bufr_new_from_file(fh)
                if bufr is None:
                    break

                ec.codes_set(bufr, "unpack", 1)

                if not ec.codes_is_defined(bufr, key):
                    logging.warning(f"Key: {key} does not exist in BUFR file")
                    return None

                # if is the first message initialise our final array
                if (msgCount == 0):
                    arr = da.from_array(ec.codes_get_array(bufr, key, float), chunks=CHUNK_SIZE)

                else:
                    tmpArr = da.from_array(ec.codes_get_array(bufr, key, float), chunks=CHUNK_SIZE)
                    arr = da.concatenate((arr, tmpArr))

                msgCount = msgCount+1
                ec.codes_release(bufr)

        if arr.size == 1:
            arr = arr[0]

        return arr

    def get_dataset(self, dataset_id, dataset_info):
        """Create dataset.

        Load data from BUFR file using the BUFR key in dataset_info
        and create the dataset with or without an AreaDefinition.

        """
        arr = self.get_array(dataset_info["key"])

        if self.with_adef and "resolution" in dataset_id:
            xarr = self.get_dataset_with_area_def(arr, dataset_id)
            # Coordinates are not relevant when returning data with an AreaDefinition
            if "coordinates" in dataset_info.keys():
                del dataset_info["coordinates"]
        else:
            if self.with_adef:
                logging.warning("Trying to use `with_area_definition=True` for a dataset without resolution. "
                                "This is not supported and the keyword will be ignored.")
            xarr = xr.DataArray(arr, dims=["y"])

        if "fill_value" in dataset_info:
            xarr = xarr.where(xarr != dataset_info["fill_value"])

        self._add_attributes(xarr, dataset_info)
        return xarr

    def get_dataset_with_area_def(self, arr, dataset_id):
        """Get dataset with an AreaDefinition."""
        if dataset_id["name"] in ["latitude", "longitude"]:
            self.__setattr__(dataset_id["name"], arr)
            xarr = xr.DataArray(arr, dims=["y"])
        else:
            lons_1d, lats_1d, data_1d = da.compute(self.longitude, self.latitude, arr)
            self._area_def = self._construct_area_def(dataset_id)
            icol, irow = self._area_def.get_array_indices_from_lonlat(lons_1d, lats_1d)

            data_2d = np.empty(self._area_def.shape)
            data_2d[:] = np.nan
            data_2d[irow.compressed(), icol.compressed()] = data_1d[~irow.mask]

            xarr = xr.DataArray(da.from_array(data_2d, CHUNK_SIZE), dims=("y", "x"))

            ntotal = len(icol)
            nvalid = len(icol.compressed())
            if nvalid < ntotal:
                logging.warning(f"{ntotal-nvalid} out of {ntotal} data points could not be put on "
                                f"the grid {self._area_def.area_id}.")

        return xarr

    def _construct_area_def(self, dataset_id):
        """Construct a standardized AreaDefinition based on satellite, instrument, resolution and sub-satellite point.

        Returns:
            A pyresample AreaDefinition object containing the area definition.

        """
        area_naming_input_dict = {"platform_name": self.platform_name[:3].lower(),
                                  "instrument_name": self.sensor_name,
                                  "resolution": dataset_id["resolution"],
                                  }

        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode(self.sensor_name, self.ssp_lon)})

        # Datasets with a segment size of 3 pixels extend outside the original SEVIRI 3km grid (with 1238 x 1238
        # segments a 3 pixels). Hence, we need to use corresponding area defintions in areas.yaml
        if self.sensor_name=="seviri" and round(dataset_id["resolution"])==9001:
            area_naming["area_id"] += "_ext"
            area_naming["description"] += " (extended outside original 3km grid)"

        # Construct AreaDefinition from standardized area definition in areas.yaml.
        stand_area_def = get_area_def(area_naming["area_id"])

        return stand_area_def

    def _add_attributes(self, xarr, dataset_info):
        """Add dataset attributes to xarray."""
        xarr.attrs["sensor"] = self.sensor_name.upper()
        xarr.attrs["platform_name"] = self.platform_name
        xarr.attrs["ssp_lon"] = self.ssp_lon
        if ("resolution" not in dataset_info) or (dataset_info["resolution"] is None):
            xarr.attrs["seg_size"] = None
            xarr.attrs["resolution"] = None
        else:
            xarr.attrs["seg_size"] = round(dataset_info["resolution"]/resolution_dict[self.sensor_name])

        xarr.attrs.update(dataset_info)



#!/usr/bin/env python
# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader configuration."""
from __future__ import annotations

import logging
import os
import warnings

import yaml
from yaml.loader import BaseLoader, FullLoader, UnsafeLoader

from satpy._config import config_search_paths, get_entry_points_config_dirs, glob_config
from satpy.readers.core.yaml_reader import load_yaml_configs as load_yaml_reader_configs

LOG = logging.getLogger(__name__)

# Old Name -> New Name
PENDING_OLD_READER_NAMES = {"fci_l1c_fdhsi": "fci_l1c_nc", "viirs_l2_cloud_mask_nc": "viirs_edr"}
OLD_READER_NAMES: dict[str, str] = {
    "slstr_l2": "ghrsst_l2",
}


def read_reader_config(config_files, loader=UnsafeLoader):
    """Read the reader `config_files` and return the extracted reader metadata."""
    reader_config = load_yaml_reader_configs(*config_files, loader=loader)
    return reader_config["reader"]


def configs_for_reader(reader=None):
    """Generate reader configuration files for one or more readers.

    Args:
        reader (Optional[str]): Yield configs only for this reader

    Returns: Generator of lists of configuration files

    """
    config_files = _get_configs(reader)

    for config_file in config_files:
        config_basename = os.path.basename(config_file)
        reader_name = os.path.splitext(config_basename)[0]
        paths = get_entry_points_config_dirs("satpy.readers")
        reader_configs = config_search_paths(
            os.path.join("readers", config_basename),
            search_dirs=paths, check_exists=True)

        if not reader_configs:
            # either the reader they asked for does not exist
            # or satpy is improperly configured and can't find its own readers
            raise ValueError("No reader named: {}".format(reader_name))

        yield reader_configs


def _get_configs(reader):
    if reader is not None:
        if not isinstance(reader, (list, tuple)):
            reader = [reader]

        reader = get_valid_reader_names(reader)
        # given a config filename or reader name
        return [r if r.endswith(".yaml") else r + ".yaml" for r in reader]

    paths = get_entry_points_config_dirs("satpy.readers")
    reader_configs = glob_config(os.path.join("readers", "*.yaml"), search_dirs=paths)
    return set(reader_configs)


def available_readers(
        as_dict: bool = False,
        yaml_loader: type[BaseLoader] | type[FullLoader] | type[UnsafeLoader] = yaml.loader.UnsafeLoader,
) -> list[str] | list[dict]:
    """Available readers based on current configuration.

    Args:
        as_dict: Optionally return reader information as a dictionary.
                        Default: False.
        yaml_loader:
            The yaml loader type. Default: ``yaml.loader.UnsafeLoader``.

    Returns:
        List of available reader names. If `as_dict` is `True` then
        a list of dictionaries including additionally reader information is returned.

    """
    readers = []
    for reader_configs in configs_for_reader():
        try:
            reader_info = read_reader_config(reader_configs, loader=yaml_loader)
        except (KeyError, IOError, yaml.YAMLError):
            LOG.debug("Could not import reader config from: %s", reader_configs)
            LOG.debug("Error loading YAML", exc_info=True)
            continue
        readers.append(reader_info if as_dict else reader_info["name"])
    if as_dict:
        readers = sorted(readers, key=lambda reader_info: reader_info["name"])
    else:
        readers = sorted(readers)
    return readers


def get_valid_reader_names(reader):
    """Check for old reader names or readers pending deprecation."""
    new_readers = []
    for reader_name in reader:
        if reader_name in OLD_READER_NAMES:
            raise ValueError(
                "Reader name '{}' has been deprecated, "
                "use '{}' instead.".format(reader_name,
                                           OLD_READER_NAMES[reader_name]))

        if reader_name in PENDING_OLD_READER_NAMES:
            new_name = PENDING_OLD_READER_NAMES[reader_name]
            warnings.warn(
                "Reader name '{}' is being deprecated and will be removed soon."
                "Please use '{}' instead.".format(reader_name, new_name),
                FutureWarning,
                stacklevel=2
            )
            new_readers.append(new_name)
        else:
            new_readers.append(reader_name)

    return new_readers



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for files produced by ESA's Ocean Color CCI project.

This reader currently supports the lat/lon gridded products and does not yet support the
products on a sinusoidal grid. The products on each of the composite periods (1, 5 and 8 day plus monthly)
are supported and both the merged product files (OC_PRODUCTS) and single product (RRS, CHLOR_A, IOP, K_490) are
supported.
"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
from pyresample import geometry

from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)


class OCCCIFileHandler(NetCDF4FileHandler):
    """File handler for Ocean Color CCI netCDF files."""

    @staticmethod
    def _parse_datetime(datestr):
        """Parse datetime."""
        return dt.datetime.strptime(datestr, "%Y%m%d%H%MZ")

    @property
    def start_time(self):
        """Get the start time."""
        return self._parse_datetime(self["/attr/time_coverage_start"])

    @property
    def end_time(self):
        """Get the end time."""
        return self._parse_datetime(self["/attr/time_coverage_end"])

    @property
    def composite_period(self):
        """Determine composite period from filename information."""
        comp1 = self.filename_info["composite_period_1"]
        comp2 = self.filename_info["composite_period_2"]
        if comp2 == "MONTHLY" and comp1 == "1M":
            return "monthly"
        elif comp1 == "1D":
            return "daily"
        elif comp1 == "5D":
            return "5-day"
        elif comp1 == "8D":
            return "8-day"
        else:
            raise ValueError(f"Unknown data compositing period: {comp1}_{comp2}")

    def _update_attrs(self, dataset, dataset_info):
        """Update dataset attributes."""
        dataset.attrs.update(self[dataset_info["nc_key"]].attrs)
        dataset.attrs.update(dataset_info)
        dataset.attrs["sensor"] = "merged"
        dataset.attrs["composite_period"] = self.composite_period
        # remove attributes from original file which don't apply anymore
        dataset.attrs.pop("nc_key")

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset."""
        dataset = da.squeeze(self[ds_info["nc_key"]])
        if "_FillValue" in dataset.attrs:
            dataset.data = da.where(dataset.data == dataset.attrs["_FillValue"], np.nan, dataset.data)
        self._update_attrs(dataset, ds_info)
        if "lat" in dataset.dims:
            dataset = dataset.rename({"lat": "y"})
        if "lon" in dataset.dims:
            dataset = dataset.rename({"lon": "x"})
        return dataset

    def get_area_def(self, dsid):
        """Get the area definition based on information in file.

        There is no area definition in the file itself, so we have to compute it
        from the metadata, which specifies the area extent and pixel resolution.
        """
        proj_param = "EPSG:4326"

        lon_res = float(self["/attr/geospatial_lon_resolution"])
        lat_res = float(self["/attr/geospatial_lat_resolution"])

        min_lon = self["/attr/geospatial_lon_min"]
        max_lon = self["/attr/geospatial_lon_max"]
        min_lat = self["/attr/geospatial_lat_min"]
        max_lat = self["/attr/geospatial_lat_max"]

        area_extent = (min_lon, min_lat, max_lon, max_lat)
        lon_size = np.round((max_lon - min_lon) / lon_res).astype(int)
        lat_size = np.round((max_lat - min_lat) / lat_res).astype(int)

        area = geometry.AreaDefinition("gridded_occci",
                                       "Full globe gridded area",
                                       "longlat",
                                       proj_param,
                                       lon_size,
                                       lat_size,
                                       area_extent)
        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for NWPSAF AAPP MAIA Cloud product.

https://nwpsaf.eu/site/software/aapp/

Documentation reference:

    [NWPSAF-MF-UD-003] DATA Formats
    [NWPSAF-MF-UD-009] MAIA version 4 Scientific User Manual

"""
import logging

import dask.array as da
import h5py
import numpy as np
from xarray import DataArray

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


class MAIAFileHandler(BaseFileHandler):
    """File handler for Maia files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(MAIAFileHandler, self).__init__(
            filename, filename_info, filetype_info)
        self.finfo = filename_info
        # set the day date part for end_time from the file name
        self.finfo["end_time"] = self.finfo["end_time"].replace(
            year=self.finfo["start_time"].year,
            month=self.finfo["start_time"].month,
            day=self.finfo["start_time"].day)
        if self.finfo["end_time"] < self.finfo["start_time"]:
            myday = self.finfo["end_time"].day
            self.finfo["end_time"] = self.finfo["end_time"].replace(
                day=myday + 1)
        self.selected = None
        self.read(self.filename)

    def read(self, filename):
        """Read the file."""
        self.h5 = h5py.File(filename, "r")
        missing = -9999.
        self.Lat = da.from_array(self.h5[u"DATA/Latitude"], chunks=CHUNK_SIZE) / 10000.
        self.Lon = da.from_array(self.h5[u"DATA/Longitude"], chunks=CHUNK_SIZE) / 10000.
        self.selected = (self.Lon > missing)
        self.file_content = {}
        for key in self.h5["DATA"].keys():
            self.file_content[key] = da.from_array(self.h5[u"DATA/" + key], chunks=CHUNK_SIZE)
        for key in self.h5[u"HEADER"].keys():
            self.file_content[key] = self.h5[u"HEADER/" + key][:]

        # Cloud Mask on pixel
        mask = 2**0 + 2**1 + 2**2
        lst = self.file_content[u"CloudMask"] & mask
        lst = lst / 2**0
        self.file_content[u"cma"] = lst

        # Cloud Mask confidence
        mask = 2**5 + 2**6
        lst = self.file_content[u"CloudMask"] & mask
        lst = lst / 2**5
        self.file_content[u"cma_conf"] = lst

        # Cloud Mask Quality
        mask = 2**3 + 2**4
        lst = self.file_content[u"CloudMask"] & mask
        lst = lst / 2**3
        self.file_content[u"cma_qual"] = lst

        # Opaque Cloud
        mask = 2**21
        lst = self.file_content[u"CloudMask"] & mask
        lst = lst / 2**21
        self.file_content[u"opaq_cloud"] = lst

        # land /water Background
        mask = 2**15 + 2**16 + 2**17
        lst = self.file_content[u"CloudMask"] & mask
        lst = lst / 2**15
        self.file_content[u"land_water_background"] = lst

        # CT (Actual CloudType)
        mask = 2**4 + 2**5 + 2**6 + 2**7 + 2**8
        classif = self.file_content[u"CloudType"] & mask
        classif = classif / 2**4
        self.file_content["ct"] = classif.astype(np.uint8)

    def get_platform(self, platform):
        """Get the platform."""
        if self.file_content["sat_id"] in (14,):
            return "viirs"
        else:
            return "avhrr"

    @property
    def start_time(self):
        """Get the start time."""
        return self.finfo["start_time"]

    @property
    def end_time(self):
        """Get the end time."""
        return self.finfo["end_time"]

    def get_dataset(self, key, info, out=None):
        """Get a dataset from the file."""
        logger.debug("Reading %s.", key["name"])
        values = self.file_content[key["name"]]
        selected = np.array(self.selected)
        if key["name"] in ("Latitude", "Longitude"):
            values = values / 10000.
        if key["name"] in ("Tsurf", "CloudTopPres", "CloudTopTemp"):
            goods = values > -9998.
            selected = np.array(selected & goods)
            if key["name"] in ("Tsurf", "Alt_surface", "CloudTopTemp"):
                values = values / 100.
            if key["name"] in ("CloudTopPres"):
                values = values / 10.
        else:
            selected = self.selected
        info.update(self.finfo)

        fill_value = np.nan

        if key["name"] == "ct":
            fill_value = 0
            info["_FillValue"] = 0
        ds = DataArray(values, dims=["y", "x"], attrs=info).where(selected, fill_value)

        # update dataset info with file_info
        return ds



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for compositors filling composites with other composites."""

import datetime as dt
import unittest
from unittest import mock

import dask
import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample import AreaDefinition

from satpy.tests.utils import CustomScheduler


class TestDayNightCompositor(unittest.TestCase):
    """Test DayNightCompositor."""

    def setUp(self):
        """Create test data."""
        bands = ["R", "G", "B"]
        start_time = dt.datetime(2018, 1, 1, 18, 0, 0)

        # RGB
        a = np.zeros((3, 2, 2), dtype=np.float32)
        a[:, 0, 0] = 0.1
        a[:, 0, 1] = 0.2
        a[:, 1, 0] = 0.3
        a[:, 1, 1] = 0.4
        a = da.from_array(a, a.shape)
        self.data_a = xr.DataArray(a, attrs={"test": "a", "start_time": start_time},
                                   coords={"bands": bands}, dims=("bands", "y", "x"))
        b = np.zeros((3, 2, 2), dtype=np.float32)
        b[:, 0, 0] = np.nan
        b[:, 0, 1] = 0.25
        b[:, 1, 0] = 0.50
        b[:, 1, 1] = 0.75
        b = da.from_array(b, b.shape)
        self.data_b = xr.DataArray(b, attrs={"test": "b", "start_time": start_time},
                                   coords={"bands": bands}, dims=("bands", "y", "x"))

        sza = np.array([[80., 86.], [94., 100.]], dtype=np.float32)
        sza = da.from_array(sza, sza.shape)
        self.sza = xr.DataArray(sza, dims=("y", "x"))

        # fake area
        my_area = AreaDefinition(
            "test", "", "",
            "+proj=longlat",
            2, 2,
            (-95.0, 40.0, -92.0, 43.0),
        )
        self.data_a.attrs["area"] = my_area
        self.data_b.attrs["area"] = my_area
        # not used except to check that it matches the data arrays
        self.sza.attrs["area"] = my_area

    def test_daynight_sza(self):
        """Test compositor with both day and night portions when SZA data is included."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="day_night")
            res = comp((self.data_a, self.data_b, self.sza))
            res = res.compute()
        expected = np.array([[0., 0.22122374], [0.5, 1.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected, rtol=1e-6)

    def test_daynight_area(self):
        """Test compositor both day and night portions when SZA data is not provided."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="day_night")
            res = comp((self.data_a, self.data_b))
            res = res.compute()
        expected_channel = np.array([[0., 0.33164983], [0.66835017, 1.]], dtype=np.float32)
        assert res.dtype == np.float32
        for i in range(3):
            np.testing.assert_allclose(res.values[i], expected_channel)

    def test_night_only_sza_with_alpha(self):
        """Test compositor with night portion with alpha band when SZA data is included."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="night_only", include_alpha=True)
            res = comp((self.data_b, self.sza))
            res = res.compute()
        expected_red_channel = np.array([[np.nan, 0.], [0.5, 1.]], dtype=np.float32)
        expected_alpha = np.array([[0., 0.3329599], [1., 1.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected_red_channel)
        np.testing.assert_allclose(res.values[-1], expected_alpha)

    def test_night_only_sza_without_alpha(self):
        """Test compositor with night portion without alpha band when SZA data is included."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="night_only", include_alpha=False)
            res = comp((self.data_a, self.sza))
            res = res.compute()
        expected = np.array([[0., 0.11042609], [0.6683502, 1.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected)
        assert "A" not in res.bands

    def test_night_only_area_with_alpha(self):
        """Test compositor with night portion with alpha band when SZA data is not provided."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="night_only", include_alpha=True)
            res = comp((self.data_b,))
            res = res.compute()
        expected_l_channel = np.array([[np.nan, 0.], [0.5, 1.]], dtype=np.float32)
        expected_alpha = np.array([[np.nan, 0.], [0., 0.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected_l_channel)
        np.testing.assert_allclose(res.values[-1], expected_alpha)

    def test_night_only_area_without_alpha(self):
        """Test compositor with night portion without alpha band when SZA data is not provided."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="night_only", include_alpha=False)
            res = comp((self.data_b,))
            res = res.compute()
        expected = np.array([[np.nan, 0.], [0., 0.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected)
        assert "A" not in res.bands

    def test_day_only_sza_with_alpha(self):
        """Test compositor with day portion with alpha band when SZA data is included."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="day_only", include_alpha=True)
            res = comp((self.data_a, self.sza))
            res = res.compute()
        expected_red_channel = np.array([[0., 0.33164983], [0.66835017, 1.]], dtype=np.float32)
        expected_alpha = np.array([[1., 0.6670401], [0., 0.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected_red_channel)
        np.testing.assert_allclose(res.values[-1], expected_alpha)

    def test_day_only_sza_without_alpha(self):
        """Test compositor with day portion without alpha band when SZA data is included."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="day_only", include_alpha=False)
            res = comp((self.data_a, self.sza))
            res = res.compute()
        expected_channel_data = np.array([[0., 0.22122373], [0., 0.]], dtype=np.float32)
        assert res.dtype == np.float32
        for i in range(3):
            np.testing.assert_allclose(res.values[i], expected_channel_data)
        assert "A" not in res.bands

    def test_day_only_area_with_alpha(self):
        """Test compositor with day portion with alpha_band when SZA data is not provided."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="day_only", include_alpha=True)
            res = comp((self.data_a,))
            res = res.compute()
        expected_l_channel = np.array([[0., 0.33164983], [0.66835017, 1.]], dtype=np.float32)
        expected_alpha = np.array([[1., 1.], [1., 1.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected_l_channel)
        np.testing.assert_allclose(res.values[-1], expected_alpha)

    def test_day_only_area_with_alpha_and_missing_data(self):
        """Test compositor with day portion with alpha_band when SZA data is not provided and there is missing data."""
        from satpy.composites.fill import DayNightCompositor

        with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
            comp = DayNightCompositor(name="dn_test", day_night="day_only", include_alpha=True)
            res = comp((self.data_b,))
            res = res.compute()
        expected_l_channel = np.array([[np.nan, 0.], [0.5, 1.]], dtype=np.float32)
        expected_alpha = np.array([[np.nan, 1.], [1., 1.]], dtype=np.float32)
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected_l_channel)
        np.testing.assert_allclose(res.values[-1], expected_alpha)

    def test_day_only_area_without_alpha(self):
        """Test compositor with day portion without alpha_band when SZA data is not provided."""
        from satpy.composites.fill import DayNightCompositor

        # with dask.config.set(scheduler=CustomScheduler(max_computes=1)):
        comp = DayNightCompositor(name="dn_test", day_night="day_only", include_alpha=False)
        res_dask = comp((self.data_a,))
        res = res_dask.compute()
        expected = np.array([[0., 0.33164983], [0.66835017, 1.]], dtype=np.float32)
        assert res_dask.dtype == res.dtype
        assert res.dtype == np.float32
        np.testing.assert_allclose(res.values[0], expected)
        assert "A" not in res.bands


class TestFillingCompositor(unittest.TestCase):
    """Test case for the filling compositor."""

    def test_fill(self):
        """Test filling."""
        from satpy.composites.fill import FillingCompositor
        comp = FillingCompositor(name="fill_test")
        filler = xr.DataArray(np.array([1, 2, 3, 4, 3, 2, 1]))
        red = xr.DataArray(np.array([1, 2, 3, np.nan, 3, 2, 1]))
        green = xr.DataArray(np.array([np.nan, 2, 3, 4, 3, 2, np.nan]))
        blue = xr.DataArray(np.array([4, 3, 2, 1, 2, 3, 4]))
        res = comp([filler, red, green, blue])
        np.testing.assert_allclose(res.sel(bands="R").data, filler.data)
        np.testing.assert_allclose(res.sel(bands="G").data, filler.data)
        np.testing.assert_allclose(res.sel(bands="B").data, blue.data)


class TestMultiFiller(unittest.TestCase):
    """Test case for the MultiFiller compositor."""

    def test_fill(self):
        """Test filling."""
        from satpy.composites.fill import MultiFiller
        comp = MultiFiller(name="fill_test")
        attrs = {"units": "K"}
        a = xr.DataArray(np.array([1, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]), attrs=attrs.copy())
        b = xr.DataArray(np.array([np.nan, 2, 3, np.nan, np.nan, np.nan, np.nan]), attrs=attrs.copy())
        c = xr.DataArray(np.array([np.nan, 22, 3, np.nan, np.nan, np.nan, 7]), attrs=attrs.copy())
        d = xr.DataArray(np.array([np.nan, np.nan, np.nan, np.nan, np.nan, 6, np.nan]), attrs=attrs.copy())
        e = xr.DataArray(np.array([np.nan, np.nan, np.nan, np.nan, 5, np.nan, np.nan]), attrs=attrs.copy())
        expected = xr.DataArray(np.array([1, 2, 3, np.nan, 5, 6, 7]))
        res = comp([a, b, c], optional_datasets=[d, e])
        np.testing.assert_allclose(res.data, expected.data)
        assert "units" in res.attrs
        assert res.attrs["units"] == "K"


def _enhance2dataset(dataset, convert_p=False):
    """Mock the enhance2dataset to return the original data."""
    return dataset


class TestBackgroundCompositor:
    """Test case for the background compositor."""

    @classmethod
    def setup_class(cls):
        """Create shared input data arrays."""
        foreground_data = {
            "L": np.array([[[1., 0.5], [0., np.nan]]]),
            "LA": np.array([[[1., 0.5], [0., np.nan]], [[0.5, 0.5], [0.5, 0.5]]]),
            "RGB": np.array([
                [[1., 0.5], [0., np.nan]],
                [[1., 0.5], [0., np.nan]],
                [[1., 0.5], [0., np.nan]]]),
            "RGBA": np.array([
                [[1., 0.5], [0., np.nan]],
                [[1., 0.5], [0., np.nan]],
                [[1., 0.5], [0., np.nan]],
                [[0.5, 0.5], [0., 0.5]]]),
        }
        cls.foreground_data = foreground_data

    @mock.patch("satpy.composites.fill.enhance2dataset", _enhance2dataset)
    @pytest.mark.parametrize(
        ("foreground_bands", "background_bands", "exp_bands", "exp_result"),
        [
            ("L", "L", "L", np.array([[1., 0.5], [0., 1.]])),
            ("L", "RGB", "RGB", np.array([
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]]])),
            ("LA", "LA", "LA", np.array([
                [[1., 0.75], [0.5, 1.]],
                [[1., 1.], [1., 1.]]])),
            ("LA", "RGB", "RGB", np.array([
                [[1., 0.75], [0.5, 1.]],
                [[1., 0.75], [0.5, 1.]],
                [[1., 0.75], [0.5, 1.]]])),
            ("RGB", "RGB", "RGB", np.array([
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]]])),
            ("RGB", "LA", "RGBA", np.array([
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]],
                [[1., 1.], [1., 1.]]])),
            ("RGB", "RGBA", "RGBA", np.array([
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]],
                [[1., 0.5], [0., 1.]],
                [[1., 1.], [1., 1.]]])),
            ("RGBA", "RGBA", "RGBA", np.array([
                [[1., 0.75], [1., 1.]],
                [[1., 0.75], [1., 1.]],
                [[1., 0.75], [1., 1.]],
                [[1., 1.], [1., 1.]]])),
            ("RGBA", "RGB", "RGB", np.array([
                [[1., 0.75], [1., 1.]],
                [[1., 0.75], [1., 1.]],
                [[1., 0.75], [1., 1.]]])),
        ]
    )
    def test_call(self, foreground_bands, background_bands, exp_bands, exp_result):
        """Test the background compositing."""
        from satpy.composites.fill import BackgroundCompositor
        comp = BackgroundCompositor("name")

        # L mode images
        foreground_data = self.foreground_data[foreground_bands]

        attrs = {"mode": foreground_bands, "area": "foo"}
        foreground = xr.DataArray(da.from_array(foreground_data),
                                  dims=("bands", "y", "x"),
                                  coords={"bands": [c for c in attrs["mode"]]},
                                  attrs=attrs)
        attrs = {"mode": background_bands, "area": "foo"}
        background = xr.DataArray(da.ones((len(background_bands), 2, 2)), dims=("bands", "y", "x"),
                                  coords={"bands": [c for c in attrs["mode"]]},
                                  attrs=attrs)

        res = comp([foreground, background])

        assert res.attrs["area"] == "foo"
        np.testing.assert_allclose(res, exp_result)
        assert res.attrs["mode"] == exp_bands

    @mock.patch("satpy.composites.fill.enhance2dataset", _enhance2dataset)
    def test_multiple_sensors(self):
        """Test the background compositing from multiple sensor data."""
        from satpy.composites.fill import BackgroundCompositor
        comp = BackgroundCompositor("name")

        # L mode images
        attrs = {"mode": "L", "area": "foo"}
        foreground_data = self.foreground_data["L"]
        foreground = xr.DataArray(da.from_array(foreground_data),
                                  dims=("bands", "y", "x"),
                                  coords={"bands": [c for c in attrs["mode"]]},
                                  attrs=attrs.copy())
        foreground.attrs["sensor"] = "abi"
        background = xr.DataArray(da.ones((1, 2, 2)), dims=("bands", "y", "x"),
                                  coords={"bands": [c for c in attrs["mode"]]},
                                  attrs=attrs.copy())
        background.attrs["sensor"] = "glm"
        res = comp([foreground, background])
        assert res.attrs["area"] == "foo"
        np.testing.assert_allclose(res, np.array([[1., 0.5], [0., 1.]]))
        assert res.attrs["mode"] == "L"
        assert res.attrs["sensor"] == {"abi", "glm"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2012-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Convert human coordinates (lon and lat) to an area definition.

Here is a usage example.

python coord2area_def.py france stere 42.0 51.5 -5.5 8.0 1.5
The arguments are "name proj min_lat max_lat min_lon max_lon resolution(km)".
The command above yelds the following result.

### +proj=stere +lat_0=46.75 +lon_0=1.25 +ellps=WGS84

france:
  description: france
  projection:
    proj: stere
    ellps: WGS84
    lat_0: 46.75
    lon_0: 1.25
  shape:
    height: 703
    width: 746
  area_extent:
    lower_left_xy: [-559750.381098, -505020.675776]
    upper_right_xy: [559750.381098, 549517.351948]


The first commented line is just a sum-up. The value of "description" can be changed to any descriptive text.

Such a custom yaml configuration can be profitably saved in a local areas.yaml configuration file that won't be
overridden by future updates of SatPy package. For that purpose the local processing script may have suitable
lines as reported below.

# set PPP_CONFIG_DIR for custom composites
import os
os.environ['PPP_CONFIG_DIR'] = '/my_local_path/for_satpy_configuration'

As a further functionality this script may give a quick display of the defined area,
provided the path for the GSHHG library is supplied via the "-s" option
and the modules PyCoast, Pillow and AggDraw have been installed.

python coord2area_def.py france stere 42.0 51.5 -5.5 8.0 1.5 -s /path/for/gshhs/library

The command above would first print the seen area definition and then launch a casual representation
of the area relying on the information about borders involved.

"""

import argparse
import sys

from pyproj import Proj

if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("name",
                        help="The name of the area.")
    parser.add_argument("proj",
                        help="The projection to use. Use proj.4 names, like 'stere', 'merc'...")
    parser.add_argument("min_lat",
                        help="The the latitude of the bottom of the area",
                        type=float)
    parser.add_argument("max_lat",
                        help="The the latitude of the top of the area",
                        type=float)
    parser.add_argument("min_lon",
                        help="The the longitude of the left of the area",
                        type=float)
    parser.add_argument("max_lon",
                        help="The the longitude of the right of the area",
                        type=float)
    parser.add_argument("resolution",
                        help="The resolution of the area (in km)",
                        type=float)
    parser.add_argument("-s", "--shapes",
                        help="Show a preview of the area using the coastlines in this directory")

    args = parser.parse_args()
    name = args.name
    proj = args.proj

    left = args.min_lon
    right = args.max_lon
    up = args.min_lat
    down = args.max_lat

    res = args.resolution * 1000

    lat_0 = (up + down) / 2
    lon_0 = (right + left) / 2

    p = Proj(proj=proj, lat_0=lat_0, lon_0=lon_0, ellps="WGS84")

    left_ex1, up_ex1 = p(left, up)
    right_ex1, up_ex2 = p(right, up)
    left_ex2, down_ex1 = p(left, down)
    right_ex2, down_ex2 = p(right, down)
    left_ex3, dummy = p(left, lat_0)
    right_ex3, dummy = p(right, lat_0)

    area_extent = (min(left_ex1, left_ex2, left_ex3),
                   min(up_ex1, up_ex2),
                   max(right_ex1, right_ex2, right_ex3),
                   max(down_ex1, down_ex2))

    xsize = int(round((area_extent[2] - area_extent[0]) / res))
    ysize = int(round((area_extent[3] - area_extent[1]) / res))

    proj4_string = "+" + \
        " +".join(("proj=" + proj + ",lat_0=" + str(lat_0) +
                   ",lon_0=" + str(lon_0) + ",ellps=WGS84").split(","))

    print("### " + proj4_string)
    print()
    print(name + ":")
    print("  description: " + name)
    print("  projection:")
    print("    proj: " + proj)
    print("    ellps: WGS84")
    print("    lat_0: " + str(lat_0))
    print("    lon_0: " + str(lon_0))
    print("  shape:")
    print("    height: " + str(ysize))
    print("    width: " + str(xsize))
    print("  area_extent:")
    print("    lower_left_xy: [%f, %f]" % (area_extent[0], area_extent[1]))
    print("    upper_right_xy: [%f, %f]" % (area_extent[2], area_extent[3]))

    if args.shapes is None:
        sys.exit(0)
    from PIL import Image
    from pycoast import ContourWriterAGG
    img = Image.new("RGB", (xsize, ysize))

    area_def = (proj4_string, area_extent)
    cw = ContourWriterAGG(args.shapes)

    cw.add_coastlines(img, (proj4_string, area_extent),
                      resolution="l", width=0.5)

    cw.add_grid(img, area_def, (10.0, 10.0), (2.0, 2.0), write_text=False, outline="white", outline_opacity=175,
                width=1.0, minor_outline="white", minor_outline_opacity=175, minor_width=0.2, minor_is_tick=False)
    img.show()



"""Interface to VIIRS L2 format.

This reader implements the support of L2 files generated using the VIIRS instrument on SNPP and NOAA satellite files.
The intent of this reader is to be able to reproduce images from L2 layers in NASA worldview with identical colormaps.

Currently a subset of four of these layers are supported
1. Deep Blue Aerosol Angstrom Exponent (Land and Ocean)
2. Clear Sky Confidence
3. Cloud Top Height
4. Deep Blue Aerosol Optical Thickness (Land and Ocean)
"""

import datetime as dt
import logging

import numpy as np

from satpy.readers.core.netcdf import NetCDF4FileHandler

LOG = logging.getLogger(__name__)


class VIIRSL2FileHandler(NetCDF4FileHandler):
    """NetCDF File Handler for VIIRS L2 Products."""
    def _parse_datetime(self, datestr):
        """Parse datetime."""
        return dt.datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S.000Z")

    @property
    def start_time(self):
        """Get start time."""
        return self._parse_datetime(self["/attr/time_coverage_start"])

    @property
    def end_time(self):
        """Get end time."""
        return self._parse_datetime(self["/attr/time_coverage_end"])

    @property
    def start_orbit_number(self):
        """Get start orbit number."""
        try:
            return int(self["/attr/orbit_number"])
        except KeyError:
            return int(self["/attr/OrbitNumber"])

    @property
    def end_orbit_number(self):
        """Get end orbit number."""
        try:
            return int(self["/attr/orbit_number"])
        except KeyError:
            return int(self["/attr/OrbitNumber"])

    @property
    def platform_name(self):
        """Get platform name."""
        try:
            res = self.get("/attr/platform", self.filename_info["platform_shortname"])
        except KeyError:
            res = "Unknown"

        return {
            "JPSS-1": "NOAA-20",
            "NP": "Suomi-NPP",
            "J1": "NOAA-20",
            "J2": "NOAA-21",
            "JPSS-2": "NOAA-21",
        }.get(res, res)

    @property
    def sensor_name(self):
        """Get sensor name."""
        return self["/attr/instrument"].lower()

    def _get_dataset_file_units(self, ds_info, var_path):
        file_units = ds_info.get("units")
        if file_units is None:
            file_units = self.get(var_path + "/attr/units")
        if file_units == "none" or file_units == "None":
            file_units = "1"
        return file_units

    def _get_dataset_valid_range(self, ds_info, var_path):
        valid_min = self.get(var_path + "/attr/valid_min")
        valid_max = self.get(var_path + "/attr/valid_max")
        if not valid_min and not valid_max:
            valid_range = self.get(var_path + "/attr/valid_range")
            if valid_range is not None:
                valid_min = valid_range[0]
                valid_max = valid_range[1]
        scale_factor = self.get(var_path + "/attr/scale_factor")
        scale_offset = self.get(var_path + "/attr/add_offset")
        return valid_min, valid_max, scale_factor, scale_offset

    def get_metadata(self, dataset_id, ds_info):
        """Get metadata."""
        var_path = ds_info.get("file_key", ds_info["name"])
        file_units = self._get_dataset_file_units(ds_info, var_path)

        # Get extra metadata
        i = getattr(self[var_path], "attrs", {})
        i.update(ds_info)
        i.update(dataset_id.to_dict())
        i.update(
            {
                "file_units": file_units,
                "platform_name": self.platform_name,
                "sensor": self.sensor_name,
                "start_orbit": self.start_orbit_number,
                "end_orbit": self.end_orbit_number,
            }
        )
        i.update(dataset_id.to_dict())
        return i

    def adjust_scaling_factors(self, factors, file_units, output_units):
        """Adjust scaling factors."""
        if factors is None or factors[0] is None:
            factors = [1, 0]
        if file_units == output_units:
            LOG.debug("File units and output units are the same (%s)", file_units)
            return factors
        factors = np.array(factors)
        return factors

    def available_datasets(self, configured_datasets=None):
        """Generate dataset info and their availablity.

        See
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for details.

        """
        for is_avail, ds_info in configured_datasets or []:
            if is_avail is not None:
                yield is_avail, ds_info
                continue
            ft_matches = self.file_type_matches(ds_info["file_type"])
            if ft_matches is None:
                yield None, ds_info
                continue
            var_path = ds_info.get("file_key", ds_info["name"])
            yield var_path in self, ds_info

    def get_dataset(self, ds_id, ds_info):
        """Get DataArray for specified dataset."""
        var_path = ds_info.get("file_key", ds_info["name"])
        metadata = self.get_metadata(ds_id, ds_info)
        (
            valid_min,
            valid_max,
            scale_factor,
            scale_offset,
        ) = self._get_dataset_valid_range(ds_info, var_path)
        data = self[var_path]
        # For aerdb Longitude and Latitude datasets have coordinates
        # This check is needed to work with yaml_reader
        if "long_name" in metadata and metadata["long_name"] == "Longitude":
            data.coords["Latitude"].attrs["standard_name"] = "latitude"
        elif "long_name" in metadata and metadata["long_name"] == "Latitude":
            data.coords["Longitude"].attrs["standard_name"] = "longitude"

        data.attrs.update(metadata)
        if valid_min is not None and valid_max is not None:
            data = data.where((data >= valid_min) & (data <= valid_max))
        factors = (scale_factor, scale_offset)
        factors = self.adjust_scaling_factors(
            factors, metadata["file_units"], ds_info.get("units")
        )
        if factors[0] != 1 or factors[1] != 0:
            data *= factors[0]
            data += factors[1]
        # rename dimensions to correspond to satpy's 'y' and 'x' standard
        if "number_of_lines" in data.dims:
            data = data.rename({"number_of_lines": "y", "number_of_pixels": "x"})
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Demo data download helper functions.

Each ``get_*`` function below downloads files to a local directory and returns
a list of paths to those files. Some (not all) functions have multiple options
for how the data is downloaded (via the ``method`` keyword argument)
including:

- gcsfs:
    Download data from a public google cloud storage bucket using the
    ``gcsfs`` package.
- unidata_thredds:
    Access data using OpenDAP or similar method from Unidata's
    public THREDDS server
    (https://thredds.unidata.ucar.edu/thredds/catalog.html).
- uwaos_thredds:
    Access data using OpenDAP or similar method from the
    University of Wisconsin - Madison's AOS department's THREDDS server.
- http:
    A last resort download method when nothing else is available of a
    tarball or zip file from one or more servers available to the Satpy
    project.
- uw_arcdata:
    A network mount available on many servers at the Space Science
    and Engineering Center (SSEC) at the University of Wisconsin - Madison.
    This is method is mainly meant when tutorials are taught at the SSEC
    using a Jupyter Hub server.

To use these functions, do:

    >>> from satpy import Scene, demo
    >>> filenames = demo.get_us_midlatitude_cyclone_abi()
    >>> scn = Scene(reader='abi_l1b', filenames=filenames)

"""

from .abi_l1b import get_hurricane_florence_abi  # noqa: F401, I001
from .abi_l1b import get_us_midlatitude_cyclone_abi  # noqa: F401
from .ahi_hsd import download_typhoon_surigae_ahi  # noqa: F401
from .fci import download_fci_test_data  # noqa: F401
from .seviri_hrit import download_seviri_hrit_20180228_1500  # noqa: F401
from .viirs_sdr import get_viirs_sdr_20170128_1229  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Advance Baseline Imager reader for the Level 1b format.

The files read by this reader are described in the official PUG document:

    https://www.goes-r.gov/users/docs/PUG-L1b-vol3.pdf

"""
import logging

import numpy as np

import satpy
from satpy.readers.core.abi import NC_ABI_BASE

logger = logging.getLogger(__name__)


class NC_ABI_L1B(NC_ABI_BASE):
    """File reader for individual ABI L1B NetCDF4 files."""

    def __init__(self, filename, filename_info, filetype_info, clip_negative_radiances=None):
        """Open the NetCDF file with xarray and prepare the Dataset for reading."""
        super().__init__(filename, filename_info, filetype_info)
        if clip_negative_radiances is None:
            clip_negative_radiances = satpy.config.get("readers.clip_negative_radiances")
        self.clip_negative_radiances = clip_negative_radiances

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading in get_dataset %s.", key["name"])

        # For raw cal, don't apply scale and offset, return raw file counts
        if key["calibration"] == "counts":
            radiances = self.nc["Rad"].copy()
            radiances = self._adjust_coords(radiances, "Rad")
        else:
            radiances = self["Rad"]

        # mapping of calibration types to calibration functions
        cal_dictionary = {
            "reflectance": self._vis_calibrate,
            "brightness_temperature": self._ir_calibrate,
            "radiance": self._rad_calibrate,
            "counts": self._raw_calibrate,
        }
        func = cal_dictionary[key["calibration"]]
        res = func(radiances)

        # convert to satpy standard units
        if res.attrs["units"] == "1" and key["calibration"] != "counts":
            res *= 100
            res.attrs["units"] = "%"

        self._adjust_attrs(res, key)
        return res

    def _adjust_attrs(self, data, key):
        data.attrs.update({"platform_name": self.platform_name,
                          "sensor": self.sensor})
        # Add orbital parameters
        projection = self.nc["goes_imager_projection"]
        data.attrs["orbital_parameters"] = {
            "projection_longitude": float(projection.attrs["longitude_of_projection_origin"]),
            "projection_latitude": float(projection.attrs["latitude_of_projection_origin"]),
            "projection_altitude": float(projection.attrs["perspective_point_height"]),
            "satellite_nominal_latitude": float(self["nominal_satellite_subpoint_lat"]),
            "satellite_nominal_longitude": float(self["nominal_satellite_subpoint_lon"]),
            "satellite_nominal_altitude": float(self["nominal_satellite_height"]) * 1000.,
            "yaw_flip": bool(self["yaw_flip_flag"]),
        }
        data.attrs.update(key.to_dict())
        # remove attributes that could be confusing later
        # if calibration type is raw counts, we leave them in
        if key["calibration"] != "counts":
            data.attrs.pop("_FillValue", None)
            data.attrs.pop("scale_factor", None)
            data.attrs.pop("add_offset", None)
        data.attrs.pop("_Unsigned", None)
        data.attrs.pop("ancillary_variables", None)  # Can't currently load DQF
        # although we could compute these, we'd have to update in calibration
        data.attrs.pop("valid_range", None)
        # add in information from the filename that may be useful to the user
        for attr in ("observation_type", "scene_abbr", "scan_mode", "platform_shortname", "suffix"):
            if attr in self.filename_info:
                data.attrs[attr] = self.filename_info[attr]
        # copy global attributes to metadata
        for attr in ("scene_id", "orbital_slot", "instrument_ID", "production_site", "timeline_ID"):
            data.attrs[attr] = self.nc.attrs.get(attr)
        # only include these if they are present
        for attr in ("fusion_args",):
            if attr in self.nc.attrs:
                data.attrs[attr] = self.nc.attrs[attr]

    def _rad_calibrate(self, data):
        """Calibrate any channel to radiances.

        This no-op method is just to keep the flow consistent -
        each valid cal type results in a calibration method call
        """
        res = data
        res.attrs = data.attrs
        return res

    def _raw_calibrate(self, data):
        """Calibrate any channel to raw counts.

        Useful for cases where a copy requires no calibration.
        """
        res = data
        res.attrs = data.attrs
        res.attrs["units"] = "1"
        res.attrs["long_name"] = "Raw Counts"
        res.attrs["standard_name"] = "counts"
        return res

    def _vis_calibrate(self, data):
        """Calibrate visible channels to reflectance."""
        solar_irradiance = self["esun"]
        esd = self["earth_sun_distance_anomaly_in_AU"]

        factor = np.pi * esd * esd / solar_irradiance

        res = data * np.float32(factor)
        res.attrs = data.attrs
        res.attrs["units"] = "1"
        res.attrs["long_name"] = "Bidirectional Reflectance"
        res.attrs["standard_name"] = "toa_bidirectional_reflectance"
        return res

    def _get_minimum_radiance(self, data):
        """Estimate minimum radiance from Rad DataArray."""
        attrs = data.attrs
        scale_factor = attrs["scale_factor"]
        add_offset = attrs["add_offset"]
        count_zero_rad = - add_offset / scale_factor
        count_pos = np.ceil(count_zero_rad)
        min_rad = count_pos * scale_factor + add_offset
        return min_rad

    def _ir_calibrate(self, data):
        """Calibrate IR channels to BT."""
        fk1 = float(self["planck_fk1"])
        fk2 = float(self["planck_fk2"])
        bc1 = float(self["planck_bc1"])
        bc2 = float(self["planck_bc2"])

        if self.clip_negative_radiances:
            min_rad = self._get_minimum_radiance(data)
            data = data.clip(min=data.dtype.type(min_rad))

        res = (fk2 / np.log(fk1 / data + 1) - bc1) / bc2
        res.attrs = data.attrs
        res.attrs["units"] = "K"
        res.attrs["long_name"] = "Brightness Temperature"
        res.attrs["standard_name"] = "toa_brightness_temperature"
        return res



# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""GCOM-C SGLI L1b reader.

GCOM-C has an imager instrument: SGLI
https://www.wmo-sat.info/oscar/instruments/view/505

Test data is available here:
https://suzaku.eorc.jaxa.jp/GCOM_C/data/product_std.html
The live data is available from here:
https://gportal.jaxa.jp/gpr/search?tab=1
And the format description is here:
https://gportal.jaxa.jp/gpr/assets/mng_upload/GCOM-C/SGLI_Level1_Product_Format_Description_en.pdf

"""

import datetime as dt
import logging

import dask.array as da
import h5py
import numpy as np
import xarray as xr
from dask.array.core import normalize_chunks

# from satpy import CHUNK_SIZE
from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)

resolutions = {"Q": 250,
               "K": 1000,
               "L": 1000}

polarization_keys = {0: "0",
                     -60: "m60",
                     60: "60"}


class HDF5SGLI(BaseFileHandler):
    """File handler for the SGLI l1b data."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the filehandler."""
        super().__init__(filename, filename_info, filetype_info)
        self.resolution = resolutions[self.filename_info["resolution"]]
        self.h5file = h5py.File(self.filename, "r")

    @property
    def start_time(self):
        """Get the start time."""
        the_time = self.h5file["Global_attributes"].attrs["Scene_start_time"].item()
        return dt.datetime.strptime(the_time.decode("ascii"), "%Y%m%d %H:%M:%S.%f")

    @property
    def end_time(self):
        """Get the end time."""
        the_time = self.h5file["Global_attributes"].attrs["Scene_end_time"].item()
        return dt.datetime.strptime(the_time.decode("ascii"), "%Y%m%d %H:%M:%S.%f")

    def get_dataset(self, key, info):
        """Get the dataset from the file."""
        if key["resolution"] != self.resolution:
            return
        file_key = info["file_key"]
        if key["name"].startswith("P"):
            file_key = file_key.format(polarization=polarization_keys[key["polarization"]])
        h5dataset = self.h5file[file_key]

        chunks = normalize_chunks(("auto", "auto"), h5dataset.shape, previous_chunks=h5dataset.chunks, dtype=np.float32)
        dataset = da.from_array(h5dataset, chunks=chunks)
        attrs = h5dataset.attrs

        dataset = xr.DataArray(dataset, attrs=attrs, dims=["y", "x"])
        dataset = self.prepare_dataset(key, dataset)

        dataset.attrs["platform_name"] = "GCOM-C1"
        dataset.attrs["sensor"] = "sgli"
        dataset.attrs["units"] = info["units"]
        dataset.attrs["standard_name"] = info["standard_name"]
        return dataset

    def prepare_dataset(self, key, dataset):
        """Prepare the dataset according to key."""
        with xr.set_options(keep_attrs=True):
            if key["name"].startswith(("VN", "SW", "P")):
                dataset = self.get_visible_dataset(key, dataset)
            elif key["name"].startswith("TI"):
                dataset = self.get_ir_dataset(key, dataset)
            elif key["name"].startswith(("longitude", "latitude")):
                dataset = self.get_lon_lats(key)
            elif "angle" in key["name"]:
                dataset = self.get_angles(key)
            else:
                raise KeyError(f"Unrecognized dataset {key['name']}")
        return dataset

    def get_visible_dataset(self, key, dataset):
        """Produce a DataArray with a visible channel data in it."""
        dataset = self.mask_to_14_bits(dataset)
        dataset = self.calibrate_vis(dataset, key["calibration"])
        return dataset

    def mask_to_14_bits(self, dataset):
        """Mask data to 14 bits."""
        return dataset & dataset.attrs["Mask"].item()

    def calibrate_vis(self, dataset, calibration):
        """Calibrate visible data."""
        attrs = dataset.attrs
        if calibration == "counts":
            return dataset
        if calibration == "reflectance":
            calibrated = (dataset * attrs["Slope_reflectance"] + attrs["Offset_reflectance"]) * 100
        elif calibration == "radiance":
            calibrated = dataset * attrs["Slope"] + attrs["Offset"]
        missing, _ = self.get_missing_and_saturated(attrs)
        return calibrated.where(dataset < missing)

    def get_missing_and_saturated(self, attrs):
        """Get the missing and saturation values."""
        missing_and_saturated = attrs["Bit00(LSB)-13"].item()
        mask_vals = missing_and_saturated.split(b"\n")[1:]
        missing = int(mask_vals[0].split(b":")[0].strip())
        saturation = int(mask_vals[1].split(b":")[0].strip())
        return missing, saturation

    def get_ir_dataset(self, key, dataset):
        """Produce a DataArray with an IR channel data in it."""
        dataset = self.mask_to_14_bits(dataset)
        dataset = self.calibrate_ir(dataset, key["calibration"])
        return dataset

    def calibrate_ir(self, dataset, calibration):
        """Calibrate IR channel."""
        attrs = dataset.attrs
        if calibration == "counts":
            return dataset
        elif calibration in ["radiance", "brightness_temperature"]:
            calibrated = dataset * attrs["Slope"] + attrs["Offset"]
            if calibration == "brightness_temperature":
                raise NotImplementedError("Cannot calibrate to brightness temperatures.")
                # from pyspectral.radiance_tb_conversion import radiance2tb
                # calibrated = radiance2tb(calibrated, attrs["Center_wavelength"] * 1e-9)
        missing, _ = self.get_missing_and_saturated(attrs)
        return calibrated.where(dataset < missing)

    def get_lon_lats(self, key):
        """Get lon/lats from the file."""
        lons = self.h5file["Geometry_data/Longitude"]
        lats = self.h5file["Geometry_data/Latitude"]
        attrs = lons.attrs
        resampling_interval = attrs["Resampling_interval"]
        if resampling_interval != 1:
            lons, lats = self.interpolate_spherical(lons, lats, resampling_interval)
        if key["name"].startswith("longitude"):
            dataset = lons
        else:
            dataset = lats
        return xr.DataArray(dataset, attrs=attrs, dims=["y", "x"])

    def interpolate_spherical(self, azimuthal_angle, polar_angle, resampling_interval):
        """Interpolate spherical coordinates."""
        from geotiepoints.geointerpolator import GeoSplineInterpolator

        full_shape = (self.h5file["Image_data"].attrs["Number_of_lines"],
                      self.h5file["Image_data"].attrs["Number_of_pixels"])

        tie_lines = np.arange(0, polar_angle.shape[0] * resampling_interval, resampling_interval)
        tie_cols = np.arange(0, polar_angle.shape[1] * resampling_interval, resampling_interval)

        interpolator = GeoSplineInterpolator((tie_lines, tie_cols), azimuthal_angle, polar_angle, kx=2, ky=2)
        new_azi, new_pol = interpolator.interpolate_to_shape(full_shape, chunks="auto")
        return new_azi, new_pol

    def get_angles(self, key):
        """Get angles from the file."""
        if "solar" in key["name"]:
            azi, zen, attrs = self.get_solar_angles()
        elif "satellite" in key["name"]:
            azi, zen, attrs = self.get_sensor_angles()
        if "azimuth" in key["name"]:
            dataset = azi
        else:
            dataset = zen
        dataset = xr.DataArray(dataset, attrs=attrs, dims=["y", "x"])
        return dataset

    def get_solar_angles(self):
        """Get the solar angles."""
        azi = self.h5file["Geometry_data/Solar_azimuth"]
        zen = self.h5file["Geometry_data/Solar_zenith"]
        attrs = zen.attrs
        azi = self.scale_array(azi)
        zen = self.scale_array(zen)
        return *self.get_full_angles(azi, zen, attrs), attrs

    def get_sensor_angles(self):
        """Get the solar angles."""
        azi = self.h5file["Geometry_data/Sensor_azimuth"]
        zen = self.h5file["Geometry_data/Sensor_zenith"]
        attrs = zen.attrs
        azi = self.scale_array(azi)
        zen = self.scale_array(zen)
        return *self.get_full_angles(azi, zen, attrs), attrs

    def scale_array(self, array):
        """Scale an array with its attributes `Slope` and `Offset` if available."""
        try:
            return array * array.attrs["Slope"] + array.attrs["Offset"]
        except KeyError:
            return array

    def get_full_angles(self, azi, zen, attrs):
        """Interpolate angle arrays."""
        resampling_interval = attrs["Resampling_interval"]
        if resampling_interval != 1:
            zen = zen[:] - 90
            new_azi, new_zen = self.interpolate_spherical(azi, zen, resampling_interval)
            return new_azi, new_zen + 90
        return azi, zen



# Copyright (c) 2017-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Enhancements."""

from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location

IMPORT_PATHS = {
    "stretch": "satpy.enhancements.contrast",
    "gamma": "satpy.enhancements.contrast",
    "invert": "satpy.enhancements.contrast",
    "piecewise_linear_stretch": "satpy.enhancements.contrast",
    "cira_stretch": "satpy.enhancements.contrast",
    "reinhard_to_srgb": "satpy.enhancements.contrast",
    "btemp_threshold": "satpy.enhancements.contrast",
    "jma_true_color_reproduction": "satpy.enhancements.ahi",
    "three_d_effect": "satpy.enhancements.convolution",
    "exclude_alpha": "satpy.enhancements.wrappers",
    "on_separate_bands": "satpy.enhancements.wrappers",
    "on_dask_array": "satpy.enhancements.wrappers",
    "using_map_blocks": "satpy.enhancements.wrappers",
    "lookup": "satpy.enhancements.colormap",
    "colorize": "satpy.enhancements.colormap",
    "palettize": "satpy.enhancements.colormap",
    "create_colormap": "satpy.enhancements.colormap",
}


def __getattr__(name: str) -> Any:
    new_module = IMPORT_PATHS.get(name)

    if new_module is None:
        raise AttributeError(f"module {__name__} has no attribute '{name}'")

    return _import_and_warn_new_location(new_module, name)
